// $Id: Client.java,v 1.7 2000/04/06 14:27:10 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.client.distributor;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.Iterator;
import org.mitre.simp.util.*;
import org.mitre.simp.client.model.*;
import org.mitre.simp.client.released.MessageAppend;
import org.mitre.simp.launcher.Status;

/**
 * SIMP Distributor for implementing group chat as a client.
 * @author John D. Ramsdell
 * @version March 2000 
 */
public class Client
    extends SimpAdapter
    implements ConnectionListener
{
    private final static boolean debug = false;
    private final static boolean showDate = true;
    private final MessageAppend transcript = new MessageAppend();
    private final Status status = new Status();
    private final Logger logger = new Logger(transcript, showDate);
    private final RunQueue q = new RunQueue();
    private final Thread t = new Thread(q);
    // This is accessed and modified only by the run queue thread.
    private final Roster roster = new Roster();
    private final ClientModel cm;
    private final Description desc;
    private final IMA me;

    /**
     * Creates a SIMP Distributor.
     */
    public Client(ClientModel cm, String font) {
	this.cm = cm;
	if (font != null)
	    transcript.setFont(Font.decode(font));

	desc = cm.getDescription();
	me = cm.getMe();
	t.start();
	q.addLogListener(logger);
	cm.addLogListener(logger);
	cm.addSimpListener(this);
	cm.addConnectionListener(this);

	JFrame frame = new JFrame("SIMP Distributor for " + me);
	frame.addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {
		    exit();
		}
	    });
	frame.setJMenuBar(createMenuBar());
	JPanel panel = new JPanel(new BorderLayout());
	JScrollPane scroller
	    = new JScrollPane(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
			      ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
	JViewport port = scroller.getViewport();
	port.add(transcript);
	panel.add(scroller);
	panel.add(status, BorderLayout.SOUTH);
	status.showStatus("Ready");
	frame.getContentPane().add(panel);
	frame.setSize(600, 400);
	frame.setVisible(true);
    }

    public void messageReceived(MessageEvent e) { 
	if (debug)
	    logger.log("Message received from " + e.getFrom());
	q.add(distributeMessage(e));
    }

    public void subscribed(SubscriberEvent e) { 
	if (debug)
	    logger.log("User " + e.getSubscriber() + " subscribed");
	q.add(addSubscriber(e.getSubscriber()));
    }

    public void unsubscribed(SubscriberEvent e) { 
	if (debug)
	    logger.log("User " + e.getSubscriber() + " unsubscribed");
	q.add(removeSubscriber(e.getSubscriber()));
    }

    private Runnable distributeMessage(final MessageEvent e) {
	return new Runnable() {
		public void run() {
		    boolean rosterChange = false;
		    IMA from = e.getFrom();
		    if (!roster.contains(from))	// message from 
			return;	// non subscriber
		    String body = from + " says:\n" + e.getBody();
		    IMA[] members = roster.toArray();
		    for (int i = 0; i < members.length; i++) {
			IMA to = members[i];
			if (!to.equals(from)) {
			    Answer ans = cm.send(to, body);
			    if (debug)
				showAnswer(ans, "message send");
			    if (ans.isError()) {
				roster.remove(to);
				rosterChange = true;
				logger.log("Dropping " + to + ": " + 
					   ans.getMessage());
				cm.dropSubscription(to);
			    }
			}
		    }
		    if (rosterChange) 
			updateDescription();
		}
	    };
    }

    private Runnable addSubscriber(final IMA subscriber) {
	return new Runnable() {
		public void run() {
		    if (roster.add(subscriber)) 
			updateDescription();
		}
	    };
    }

    private Runnable removeSubscriber(final IMA subscriber) {
	return new Runnable() {
		public void run() {
		    if (roster.remove(subscriber)) 
			updateDescription();
		}
	    };
    }

    private void updateDescription() {
	StringBuffer sb = new StringBuffer("Members:");
	if (roster.isEmpty())
	    sb.append(" none");
	else {
	    Iterator it = roster.iterator();
	    while (it.hasNext()) {
		IMA place = (IMA)it.next();
		sb.append(' ');
		sb.append(place.toString());
	    }
	}
	String message = sb.toString();
	logger.log(message);
	Answer ans = desc.setDescription(message);
	if (debug)
	    showAnswer(ans, "description update");
	if (ans.isError())
	    logger.log("Update description: " + ans.getMessage());
    }	

    public void connectionEstablished(ConnectionEvent e) { 
	status.showStatus("Connection established");
    }

    public void connectionClosed(ConnectionEvent e) { 
	status.showStatus("Connection closed");
    }

    private void exit() {
	cm.logout();
	t.interrupt();
	System.exit(0);
    }

    private JMenuBar createMenuBar() {
	JMenuBar mb = new JMenuBar();
	JMenu file = new JMenu("File");
	file.setMnemonic('F');

	mb.add(file);

	JMenuItem mi = new JMenuItem("Exit", 'x');
	mi.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    exit();
		}
	    });
	file.add(mi);
	return mb;
    }

    private void showAnswer(Answer ans, String where) {
	String message = "nswer in " + where + ": " + ans.getMessage();
	if (ans.isError()) 
	    message = "Error a" + message;
	else
	    message = "A" + message;
	logger.log(message);
    }
}
