// $Id: BuddyList.java,v 1.40 2000/04/14 18:36:39 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.client.model;

import java.util.*;
import java.net.*;
import java.io.IOException;
import javax.swing.SwingUtilities;
import org.mitre.simp.util.*;

/**
 * A buddy list is a collection of groups, and a group is a named set
 * of buddies.  Each buddy is identified by an IMA.  The buddy list
 * object maintains a buddy list on the client and the server.
 * Changes to the buddy list are transmitted to the server and
 * components of the client can be notified of the changes by
 * registering a BuddyListListener.  The events are delivered by the AWT
 * event thread.
 * @author John D. Ramsdell
 * @author Galen B. Williamson
 * @version February 2000 
 */
public class BuddyList
{
    // maps group names to group lists and
    // maps IMAs to sets of group names
    private volatile GroupTable groupTable = new GroupTable();

    // run queue for buddy list modifiers
    private final RunQueue q = new RunQueue();
    private final Thread t = new Thread(q);

    // The tracker periodically subscribes to all in the buddy group table.
    private final BuddyTracker tracker;

    private final ClientModel cm;
    
    BuddyList(final ClientModel cm) {
	this.cm = cm;
	tracker = new BuddyTracker(cm, this);
	q.addLogListener(new LogListener() {
		public void messageGenerated(LogEvent e) {
		    cm.log(e.getMessage(), e.getThrowable());
		}
	    });
	t.start();
    }

    void init() {
	ProfileAnswer pa = cm.getProfile();
	if (pa.isError()) {
	    cm.log("buddy list cannot get profile -- buddy tracking disable");
	    return;
	}
	XmlProperties profile = pa.getProfile();
	String buddies = profile.getProperty(Protocol.BUDDIES);
	if (buddies == null) return;
	// load the buddy table
	XmlProperties buddyTable = new XmlProperties();
	try {
	    buddyTable.load(buddies);
	}
	catch (IOException iox) {
	    cm.log("format error in buddy list");
	    return;
	}

	// put the buddy list into the group table
	Enumeration vals = buddyTable.keys();
	while (vals.hasMoreElements()) {
	    String group = (String)vals.nextElement();
	    GroupList gl = new GroupList(group);
	    groupAdded(group);
	    String members = buddyTable.getProperty(group);
	    StringTokenizer t = new StringTokenizer(members);
	    while (t.hasMoreTokens()) {
		String buddy = t.nextToken();
		if (!IMA.isIMA(buddy)) {
		    cm.log("buddy syntax error: " + buddy);
		    continue;
		}
		try {
		    IMA to = new IMA(buddy);
		    gl.add(to);
		    buddyAdded(to, group);
		}
		catch (UnknownHostException ex) {
		    String msg = "buddy host is unknown: " + buddy;
		    cm.log(msg);
		}
	    }
	    groupTable.put(group, gl);
	}
	// start the tracker
	new Thread(tracker).start();
    }

    // buddy queries

    /**
     * Returns the set of groups
     */
    public Set getGroups() {
	return groupTable.keySet();
    }

    /**
     * Returns the set of groups of which the buddy is a member.
     * Returns null if the buddy is not in any group.
     */
    public Set getGroups(IMA buddy) {
	HashSet s = groupTable.getGroups(buddy);
	return s != null
	    ? (Set)s.clone()
	    : null;
    }

    /**
     * Does group exist?
     */
    public boolean groupExists(String group) {
	return null != groupTable.get(group);
    }

    /**
     * Returns the buddies in a group.
     */
    public IMA[] getBuddies(String group) {
	GroupList gl = groupTable.get(group);
	if (gl == null)
	    return new IMA[0];
	else
	    return gl.getMembers();
    }

    // This is for the tracker
    GroupTable getGroupTable() {
	return groupTable;
    }

    /**
     * Resume the buddy list tracker.  The tracker checks subscriptions
     * and subscribes to buddies as needed.  The tracker makes this check
     * periodically.  This method forces the performance of the check
     * immediately.
     */
    public void resumeTracker() {
	tracker.resume();
    }

    // buddy list modifiers

   /**
     * Add a buddy to a group.
     * An error occurs if the group is missing, 
     * the buddy is already there, 
     * or if the server has a problem writing the new buddy list.
     */
    public void addBuddy(final AnswerClosure closure, 
			 final IMA buddy, final String group) {
	q.add(new Runnable() {
		public void run() {
		    Answer ans = reallyAddBuddy(buddy, group);
		    SwingUtilities.invokeLater(closure.invoke(ans));
		}
	    });
    }

    /**
     * Remove a buddy from a group.
     * An error occurs if the group is missing, 
     * the buddy is not in the group,
     * or if the server has a problem writing the new buddy list.
     */
    public void removeBuddy(final AnswerClosure closure, 
			    final IMA buddy, final String group) {
	q.add(new Runnable() {
		public void run() {
		    Answer ans = reallyRemoveBuddy(buddy, group);
		    SwingUtilities.invokeLater(closure.invoke(ans));
		}
	    });
    }

    /**
     * Add a group.
     * An error occurs if the group exists,
     * or if the server has a problem writing the new buddy list.
     */
    public void addGroup(final AnswerClosure closure, 
			 final String group) {
	q.add(new Runnable() {
		public void run() {
		    Answer ans = reallyAddGroup(group);
		    SwingUtilities.invokeLater(closure.invoke(ans));
		}
	    });
    }

    /**
     * Remove a group.
     * An error occurs if the group is missing,
     * or if the server has a problem writing the new buddy list.
     */
    public void removeGroup(final AnswerClosure closure, 
			    final String group) {
	q.add(new Runnable() {
		public void run() {
		    Answer ans = reallyRemoveGroup(group);
		    SwingUtilities.invokeLater(closure.invoke(ans));
		}
	    });
    }

    /**
     * Rename a group.
     * An error occurs if the new group exists,
     * the old group does not exist,
     * or if the server has a problem writing the new buddy list.
     */
    public void renameGroup(final AnswerClosure closure, 
			    final String to, final String from) {
	q.add(new Runnable() {
		public void run() {
		    Answer ans = reallyRenameGroup(to, from);
		    SwingUtilities.invokeLater(closure.invoke(ans));
		}
	    });
    }

    /**
     * Add a buddy to a group.
     * An error occurs if the group is missing, 
     * the buddy is already there, 
     * or if the server has a problem writing the new buddy list.
     * This method should only be used within the buddy list due to 
     * synchronization errors.
     */
    Answer reallyAddBuddy(IMA buddy, String group) {
	GroupTable gt = (GroupTable)groupTable.clone();
	GroupList gl = gt.get(group);
	if (gl == null)
	    return new Answer(true, "No such group: "+group);
	if (gl.isMember(buddy))
	    return new Answer(true, "Buddy is already in group");
	gl = (GroupList)gl.clone();
	gl.add(buddy);
	gt.put(group, gl);
	Answer a = saveGroupTable(gt, "Buddy added");
	if (!a.isError()) {
	    groupTable = gt;
	    buddyAdded(buddy, group);
	    tracker.resume();
	}
	return a; 
    }

    private Answer saveGroupTable(GroupTable gt, String message) {
	XmlProperties buddyTable = gt.asXp();
	ProfileAnswer pa = cm.getProfile();
	if (pa.isError())
	    return pa;
	XmlProperties profile = pa.getProfile();
	try {
	    profile.setProperty(Protocol.BUDDIES, buddyTable.asString());
	} catch (IOException iox) { 
	    // Ignoring this because it could only happen if
	    // UTF8 is an unsupported encoding, in which case
	    // the whole dern thing probably won't work anyway.
	    cm.log("save group table: "+iox.toString());
	}
	Answer answer = cm.setProfile(profile);
	if (answer.isError())
	    return answer;
	return new Answer(false, message);
    }

    /**
     * Remove a buddy from a group.
     * An error occurs if the group is missing, 
     * the buddy is not in the group,
     * or if the server has a problem writing the new buddy list.
     * This method should only be used within the buddy list due to 
     * synchronization errors.
     */
    Answer reallyRemoveBuddy(IMA buddy, String group) {
	GroupTable gt = (GroupTable)groupTable.clone();
	GroupList gl = gt.get(group);
	if (gl == null)
	    return new Answer(true, "No such group: "+group);
	if (!gl.isMember(buddy))
	    return new Answer(true, "Buddy is not in group");
	gl = (GroupList)gl.clone();
	gl.remove(buddy);
	gt.put(group, gl);
	Answer a = saveGroupTable(gt, "Buddy added");
	if (!a.isError()) {
	    groupTable = gt;
	    buddyRemoved(buddy, group);
	    if (gt.getGroups(buddy) == null) {
		Answer ans = cm.unsubscribeFrom(buddy);
		if (ans.isError()) {
		    cm.log("While unsubscribing from " + buddy  +
			   ": " + ans.getMessage());
		}
	    }
	}
	return a; 
    }

    /**
     * Add a group.
     * An error occurs if the group exists,
     * or if the server has a problem writing the new buddy list.
     * This method should only be used within the buddy list due to 
     * synchronization errors.
     */
    Answer reallyAddGroup(String group) {
	if (group == null || "".equals(group)) 
	    return new Answer(true, "Bad group name");
	GroupTable gt = (GroupTable)groupTable.clone();
	GroupList gl = gt.get(group);
	if (gl != null) 
	    return new Answer(true, "Group exists");
	gt.put(group, new GroupList(group));
	Answer a = saveGroupTable(gt, "Group added");
	if (!a.isError()) {
	    groupTable = gt;
	    groupAdded(group);
	}
	return a; 
    }

    /**
     * Remove a group.
     * An error occurs if the group is missing,
     * or if the server has a problem writing the new buddy list.
     * This method should only be used within the buddy list due to 
     * synchronization errors.
     */
    Answer reallyRemoveGroup(String group) {
	GroupTable gt = (GroupTable)groupTable.clone();
	GroupList gl = gt.get(group);
	if (gl == null) 
	    return new Answer(true, "Group does not exist");
	gt.remove(group);
	Answer a = saveGroupTable(gt, "Group removed");
	if (!a.isError()) {
	    groupTable = gt;
	    groupRemoved(group);
	}
	return a; 
    }

    /**
     * Rename a group.
     * An error occurs if the new group exists,
     * the old group does not exist,
     * or if the server has a problem writing the new buddy list.
     * This method should only be used within the buddy list due to 
     * synchronization errors.
     */
    Answer reallyRenameGroup(String to, String from) {
	if (to == null 
	    || from == null
	    || "".equals(to) 
	    || to.equals(from)) 
	    return new Answer(true, "Bad group name");
	GroupTable gt = (GroupTable)groupTable.clone();
	GroupList ngl = gt.get(to);
	if (ngl != null) 
	    return new Answer(true, "Group exists");
	GroupList ogl = gt.remove(from);
	if (ogl == null) 
	    return new Answer(true, "Group does not exist");
	gt.put(to, (GroupList)ogl.clone());
	Answer a = saveGroupTable(gt, "Group renamed");
	if (!a.isError()) {
	    groupTable = gt;
	    groupRenamed(new GroupRenameEvent(this, to, from));
	}
	return a; 
    }

    /** 
     * Register conversationalist and receive presence info.
     * The conversationalist need not be in the buddy list.
     */
    public Answer addConversationBuddy(IMA conversationalist) {
	if (groupTable.getGroups(conversationalist) == null)
	    return cm.subscribeTo(conversationalist);
	else
	    return new Answer(false, "okay");
    }

    /** 
     * Unregister conversationalist.
     */
    public Answer removeConversationBuddy(IMA conversationalist) {
	if (groupTable.getGroups(conversationalist) == null)
	    return cm.unsubscribeFrom(conversationalist);
	else
	    return new Answer(false, "okay");
    }

    // Buddy listener

    private Vector buddyListeners = new Vector();

    public synchronized void addBuddyListListener(BuddyListListener el) {
	buddyListeners.add(el);
    }

    public synchronized void removeBuddyListListener(BuddyListListener el) {
	buddyListeners.remove(el);
    }

    private void buddyAdded(IMA buddy, String group) {
	GroupEvent e = new GroupEvent(this, buddy, group);
	groupChanged(e, true);
    }

    private void buddyRemoved(IMA buddy, String group) {
	GroupEvent e = new GroupEvent(this, buddy, group);
	groupChanged(e, false);
    }

    private void groupChanged(GroupEvent e, boolean create) {
	Vector l;
	synchronized (this) { l = (Vector)buddyListeners.clone(); }
	deliverGroupEvent(l, e, create);
    }

    private void deliverGroupEvent(final Vector l, 
				   final GroupEvent e,
				   final boolean create) {
	if (SwingUtilities.isEventDispatchThread()) {
	    for (int i = 0; i < l.size(); i++) {
		if (create)
		    ((BuddyListListener)l.elementAt(i)).buddyAdded(e);
		else
		    ((BuddyListListener)l.elementAt(i)).buddyRemoved(e);
	    }
	}
	else {
	    SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
			deliverGroupEvent(l, e, create);
		    }
		});
	}
    }

    private void groupRenamed(GroupRenameEvent e) {
	Vector l;
	synchronized (this) { l = (Vector)buddyListeners.clone(); }
	deliverGroupRenameEvent(l, e);
    }

    private void deliverGroupRenameEvent(final Vector l, 
					 final GroupRenameEvent e) {
	if (SwingUtilities.isEventDispatchThread()) {
	    for (int i = 0; i < l.size(); i++) {
		((BuddyListListener)l.elementAt(i)).groupRenamed(e);
	    }
	}
	else {
	    SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
			deliverGroupRenameEvent(l, e);
		    }
		});
	}
    }

    private void groupAdded(String group) {
	BuddyListEvent e = new BuddyListEvent(this, group);
	buddyListChanged(e, true);
    }

    private void groupRemoved(String group) {
	BuddyListEvent e = new BuddyListEvent(this, group);
	buddyListChanged(e, false);
    }

    private void buddyListChanged(BuddyListEvent e, boolean create) {
	Vector l;
	synchronized (this) { l = (Vector)buddyListeners.clone(); }
	deliverBuddyListEvent(l, e, create);
    }

    private void deliverBuddyListEvent(final Vector l, 
				       final BuddyListEvent e,
				       final boolean create) {
	if (SwingUtilities.isEventDispatchThread()) {
	    for (int i = 0; i < l.size(); i++) {
		if (create)
		    ((BuddyListListener)l.elementAt(i)).groupAdded(e);
		else
		    ((BuddyListListener)l.elementAt(i)).groupRemoved(e);
	    }
	}
	else {
	    SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
			deliverBuddyListEvent(l, e, create);
		    }
		});
	}
    }
}
