// $Id: ClientModel.java,v 1.50 2000/04/13 11:22:16 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.client.model;

import java.util.*;
import java.io.*;
import java.net.*;
import javax.swing.SwingUtilities;
import org.mitre.simp.util.*;
import org.mitre.simp.launcher.LoginResult;

/**
 * The client model handles the details of the SIMP protocol.  In
 * addition to providing methods for sending messages and requesting
 * presence information, it provides methods for manipulating the
 * user's data stored on the server.  It also generates a variety of
 * events in response to information send by the server.  Events set
 * to SimpListeners are distinguished by the fact that they are
 * delivered on the AWT thread.  By listening to connection events,
 * one can find out when the model has finished initializing.
 * @author John D. Ramsdell
 * @author Galen B. Williamson
 * @version February 2000 
 */
public class ClientModel
    extends LogGenerator
{
    final static long LEASE_DURATION = 30 * 60 * 1000;

    private boolean debug = false;
    private final IMA me;
    private final XpSocket xps;
    private XmlProperties profile;
    private ACL acl;
    private final Endpoint ep;
    private final BuddyList bl;
    private final Description desc;

    /**
     * Creates a client model.
     */    
    public ClientModel(IMA me, XpSocket xps, XmlProperties profile) {
	this.me = me;
	this.xps = xps;
	this.profile = profile;
	xps.addLogListener(new LogListener() {
		public void messageGenerated(LogEvent e) {
		    log(e.getMessage(), e.getThrowable());
		}
	    });		    
	ep = new Endpoint(this, xps);
	bl = new BuddyList(this);
	String msg = null;
	if (profile != null)
	    msg = profile.getProperty(Protocol.MESSAGE);
	desc = new Description(this, msg);
    }

    // The following methods are (only?) used by the Endpoint
    void reply(String id, XmlProperties resp) {
	resp.setProperty(SocketPost.RESPONSE_KEY, id);
	try {
	    xps.write(resp);
	}
	catch (IOException iox) {
	    log("Reply: " + iox.toString());
	}
    }

    boolean noteMessage(XmlProperties xp) {
	try {
	    String toS = xp.getProperty(Protocol.TO);
	    String fromS = xp.getProperty(Protocol.FROM);
	    String replyToS = xp.getProperty(Protocol.REPLY_TO);
	    String type = xp.getProperty(Protocol.TYPE);
	    String body = xp.getProperty(Protocol.BODY);
	    if (toS == null || fromS == null 
		|| type == null || body == null)
		return false;
	    if (!IMA.isIMA(toS) || !IMA.isIMA(fromS))
		return false;
	    IMA to = new IMA(toS);
	    IMA from = new IMA(fromS);
	    IMA replyTo = null;
	    if (replyToS != null) {
		if (!IMA.isIMA(replyToS))
		    return false;
		replyTo = new IMA(replyToS);
	    }
	    noteMessageEvent(to, from, replyTo, type, body);
	    return true;
	}
	catch (UnknownHostException e) {
	    log("Note Message: " + e.toString());
	    return false;
	}
    }

    boolean noteInfo(XmlProperties xp) {
	if (!isFromServer(xp))
	    return false;
	Info info = makeInfo(xp);
	if (xp == null)
	    return false;
	noteInfoEvent(info);
	return true;
    }

    private boolean isFromServer(XmlProperties xp) {
	String fromS = xp.getProperty(Protocol.FROM);
	if (IMA.isIMA(fromS)) {
	    try {
		IMA from = new IMA(fromS);
		return Protocol.SERVER_USER.equals(from.getUser());
	    }
	    catch (UnknownHostException e) {
	    }
	}
	return false;
    }
	
    private Info makeInfo(XmlProperties xp) {
	try {
	    String regardingS = xp.getProperty(Protocol.REGARDING);
	    String status = xp.getProperty(Protocol.STATUS);
	    String onlineS = xp.getProperty(Protocol.ON_SINCE);
	    String message = xp.getProperty(Protocol.MESSAGE);
	    if (regardingS == null || status == null)
		return null;
	    if (!IMA.isIMA(regardingS))
		return null;
	    IMA regarding = new IMA(regardingS);
	    DateTime online = null;
	    if (onlineS != null)
		online = new DateTime(onlineS);
	    return new Info(regarding, status, online,  message);
	}
	catch (UnknownHostException e) {
	    log(e.toString());
	    return null;
	}
    }

    void noteSubscriber(XmlProperties xp) {
	noteSubscription(xp, true);
    }

    void noteUnsubscriber(XmlProperties xp) {
	noteSubscription(xp, false);
    }

    private final Hashtable subscriberTable = new Hashtable();

    void noteSubscription(XmlProperties xp, boolean create) {
	String regarding = xp.getProperty(Protocol.SUBSCRIBER);
	if (!IMA.isIMA(regarding))
	    return;
	try {
	    IMA ima = new IMA(regarding);
	    if (create) {
		if (subscriberTable.put(ima, ima) != null)
		    return;
	    }
	    else if (subscriberTable.remove(ima) == null)
		return;
	    noteSubscriberEvent(ima, create);
	}
	catch (UnknownHostException uhe) { }
    }

    boolean noteSubscriptionEnd(XmlProperties xp) {
	if (!isFromServer(xp))
	    return false;
	Info info = makeInfo(xp);
	if (xp == null)
	    return false;
	noteInfoEvent(info);
	// If you trust the message - do this:
	removeLease(info.getRegarding());
	return true;
    }

    // end of endpoint support section

    void init() {
	if (profile == null) {
	    log("No profile");
	    return;
	}
	bl.init();
	Thread t = new Thread(new Runnable() {
		public void run() {
		    noteConnectionEvent("Connection Established", 
					true);
		}
	    });
	t.start();
    }

    /**
     * Closes the connection to the server, thus ending a session.
     */
    public void logout() {
	try {
	    xps.close();
	}
	catch (IOException iox) { }
    }

    // This object is used to synchronized subscribe and 
    // unsubscribe requests.
    private Object subscriptionLock = new Object();

    Answer subscribeTo(IMA to) {
	synchronized (subscriptionLock) {
	    String msg = null;
	    // Check lease to see if you already have one.
	    Lease lease = getLease(to);
	    if (lease != null) {
		long now = System.currentTimeMillis();
		if (now + LEASE_DURATION < lease.getDone())
		    return new Answer(false, "lease already in hand");
	    }	    
	    if (debug) 
		log("Subscribing to: " + to);
	    try {
		XmlProperties req 
		    = Protocol.subscribe(to, me, 2 * LEASE_DURATION);
		XmlProperties resp = ep.post(req);
		String action = resp.getProperty(Protocol.ACTION);
		String message = resp.getProperty(Protocol.MESSAGE,
						  "Subscribe to");
		if (!Protocol.DESCRIBE.equals(action)) {
		    log("Error subscribing to buddy: " + to);
		    msg = "Error subscribing to buddy: " + message;
		}
		else {
		    String opaque = resp.getProperty(Protocol.OPAQUE);
		    long duration = resp.getPropertyAsLong(Protocol.DURATION, 0);
		    Info info = makeInfo(resp);
		    if (info == null || opaque == null || duration == 0) 
			msg = "Format error subscribing to buddy: " + message;
		    else {
			addLease(info.getRegarding(), opaque, duration);
			noteInfoEvent(info);
			return new Answer(false, message);
		    }
		}
	    }
	    catch (IOException ex) {
		msg = "error while sending to " + to + ": " + ex.getMessage();
		log(msg);
	    }
	    return new Answer(true, msg);
	}
    }

    Answer unsubscribeFrom(IMA to) {
	synchronized (subscriptionLock) {
	    String msg = null;
	    Lease lease = getLease(to);
	    if (lease == null)
		msg = "No lease for " + to.toString();
	    else {
		try {
		    XmlProperties req = Protocol.unsubscribe(to, me, 
							     lease.getOpaque());
		    XmlProperties resp = ep.post(req);
		    String action = resp.getProperty(Protocol.ACTION);
		    String message = resp.getProperty(Protocol.MESSAGE,
						      "Unsubscribe from");
		    if (!Protocol.OKAY.equals(action)) {
			log("Error unsubscribing from buddy: " + to);
			msg = "Error unsubscribing from buddy: " + message;
		    }
		    else {
			removeLease(to);
			return new Answer(false, message);
		    }
		}
		catch (IOException ex) {
		    msg = "error while sending to " + to + ": " + ex.getMessage();
		    log(msg);
		}
	    }
	    return new Answer(true, msg);
	}
    }

    // lease table

    // maps IMAs to Leases
    private final Hashtable leaseTable = new Hashtable();

    // Do you really need to store the place in the lease?
    private void addLease(IMA place, String opaque, long duration) {
	Lease lease = new Lease(opaque, duration);
	if (leaseTable.put(place, lease) == null)
	    noteLeaseEvent(place, true);
    }

    private void removeLease(IMA place) {
	removeInfoTablePlace(place);
	if (leaseTable.remove(place) != null)
	    noteLeaseEvent(place, false);
    }

    Lease getLease(IMA place) {
	synchronized (leaseTable) {
	    Lease lease = (Lease)leaseTable.get(place);
	    if (lease != null) {
		long now = System.currentTimeMillis();
		if (now < lease.getDone())
		    return lease;
		removeLease(place);
	    }
	}
	return null;
    }

    /**
     * Sends a plain text message to a recipient.
     */
    public Answer send(IMA to, String body) {
	return send(to, "text/plain", body);
    }

    /**
     * Sends a typed message to a recipient.
     */
    public Answer send(IMA to, String type, String body) {
	XmlProperties xp = Protocol.send(to, me, null, type, body);
	try {
	    xp = ep.post(xp);
	}
	catch (IOException iox) {
	    log("Send: "+iox.toString());
	    return new Answer(true, "error during send");
	}
	String action = xp.getProperty(Protocol.ACTION);
	return new Answer(!Protocol.OKAY.equals(action),
			  xp.getProperty(Protocol.MESSAGE));
    }

    /**
     * Fetches the presence information of a user.
     */
    public InfoAnswer fetch(IMA to) {
	XmlProperties xp = Protocol.fetch(to, me);
	try {
	    xp = ep.post(xp);
	}
	catch (IOException iox) {
	    log("Fetch: "+iox.toString());
	    return new InfoAnswer("error during fetch");
	}
	String action = xp.getProperty(Protocol.ACTION);
	if (!Protocol.DESCRIBE.equals(action)) {
	    String msg = xp.getProperty(Protocol.MESSAGE);
	    String err = "error during fetch";
	    if (msg != null)
		err += ": "+msg;
	    log("Fetch: "+err);
	    return new InfoAnswer(err);
	}
	Info info = makeInfo(xp);
	if (info != null) {
	    noteInfoEvent(info);
    	    return new InfoAnswer(info);
	}
	else
	    return new InfoAnswer("description garbled");
    }

    /**
     * Inquire about the server providing presence info for the IMA to.
     */
    public Answer inquire(IMA to) {
	XmlProperties xp = Protocol.inquire(to, me);
	try {
	    xp = ep.post(xp);
	}
	catch (IOException iox) {
	    log("Inquire: "+iox.toString());
	    return new Answer(true, "error during inquire");
	}
	String action = xp.getProperty(Protocol.ACTION);
	return new Answer(!Protocol.OKAY.equals(action),
			  xp.getProperty(Protocol.MESSAGE));
    }

    /**
     * Drops a subscriber.
     */
    public Answer dropSubscription(IMA subscriber) {
	XmlProperties xp = Protocol.dropSubscription(subscriber);
	try {
	    xp = ep.post(xp);
	}
	catch (IOException iox) {
	    log("Send: "+iox.toString());
	    return new Answer(true, "error during drop subscriber");
	}
	String action = xp.getProperty(Protocol.ACTION);
	return new Answer(!Protocol.OKAY.equals(action),
			  xp.getProperty(Protocol.MESSAGE));
    }

    /**
     * Returns the object that manages the buddy list of the
     * user of the client.
     */
    public BuddyList getBuddyList() {
	return bl;
    }

    /**
     * Returns the object that manages the description of the
     * user of the client.
     */
    public Description getDescription() {
	return desc;
    }

    // maps IMAs to Infos
    // has an entry only when the client has a subscription
    private final Hashtable infoTable = new Hashtable();

    private void updateInfoTable(Info info) {
	// need to filter out useless info events
	IMA regarding = info.getRegarding();
	if (getLease(regarding) == null)
	    return;
	Info old = (Info)infoTable.put(regarding, info);
	if (!info.isInfoSame(old))
	    buddyInfoChanged(info);
    }

    private void removeInfoTablePlace(IMA place) {
	infoTable.remove(place);
    }


    /**
     * Returns the latest presence info for a buddy.
     */
    public Info getInfo(IMA buddy) {
	return (Info)infoTable.get(buddy);
    }

    /**
     * Returns my IMA.
     */
    public IMA getMe() {
	return me;
    }

    /**
     * Is the client model in debugging mode.
     */
    public boolean isDebugging() {
	return debug;
    }

    /**
     * Set debugging mode.  When enabled, more detailed 
     * output is placed in the log.
     */
    public void setDebugging(boolean debug) {
	this.debug = debug;
    }

    /**
     * Returns the user's profile.
     */
    public ProfileAnswer getProfile() {
	return new ProfileAnswer(profile == null, "", profile);
    }
    
    /**
     * Sets the user's profile.
     */
    public Answer setProfile(XmlProperties profile) {
	XmlProperties xp = Protocol.setProfile(profile);
	try {
	    xp = ep.post(xp);
	}
	catch (IOException iox) {
	    log("Set profile: "+iox.toString());
	    return new Answer(true, "error during set profile");
	}
	String action = xp.getProperty(Protocol.ACTION);
	boolean okay = Protocol.OKAY.equals(action);
	String msg = xp.getProperty(Protocol.MESSAGE);
	if (okay) {
	    this.profile = profile;
	    noteProfile(profile);
	} else {
	    log("Set profile: "+msg);
	}
	return new Answer(!okay, msg);
    }

    private ACLAnswer aclAnswerError(String msg) {
	return new ACLAnswer(true, msg, null);
    }

    /**
     * Returns the user's access control list.
     */
    public ACLAnswer getACL() {
	if (acl == null) {
	    XmlProperties xp = Protocol.getACL();
	    try {
		xp = ep.post(xp);
	    }
	    catch (IOException iox) {
		log("Get ACL: "+iox.toString());
		return aclAnswerError("error during get ACL");
	    }
	    String action = xp.remove(Protocol.ACTION);
	    boolean okay = Protocol.SELF.equals(action);
	    String msg = xp.getProperty(Protocol.MESSAGE);
	    if (okay) {
		ACL a = ACL.encode(xp);
		if (a != null) {
		    acl = a;
		    noteACL(a);
		}
		else {
		    okay = false;
		    log("format error in acl");
		} 
	    }
	    else {
		log("Get ACL: "+msg);
	    }
	    return new ACLAnswer(!okay, msg, acl);

	}
	else 
	    return new ACLAnswer(false, "", acl);
    }

    /**
     * Sets the user's access control list.
     */
    public Answer setACL(ACL acl) {
	if (acl == null)
	    return new Answer(true, "null acl");
	XmlProperties xp = Protocol.setACL(acl.getXp());
	try {
	    xp = ep.post(xp);
	}
	catch (IOException iox) {
	    log("Set ACL: "+iox.toString());
	    return new Answer(true, "error during set ACL");
	}
	String action = xp.getProperty(Protocol.ACTION);
	boolean okay = Protocol.OKAY.equals(action);
	String msg = xp.getProperty(Protocol.MESSAGE);
	if (okay) {
	    this.acl = acl;
	    noteACL(acl);
	} else {
	    log("Set ACL: "+msg);
	}
	return new Answer(!okay, msg);
    }

    /* Event generators */

    // client model listener

    private Vector simpListeners = new Vector();

    public synchronized void addSimpListener(SimpListener el) {
	simpListeners.add(el);
    }

    public synchronized void removeSimpListener(SimpListener el) {
	simpListeners.remove(el);
    }

    // Only used during debugging.
    private String infoString(Info info) {
	return 
	    info.getRegarding().toString() +
	    " " + info.getStatus() +
	    " " + info.getMessage();
    }

    private void noteInfoEvent(Info info) {
	if (debug)
	    log("Note info for " + infoString(info));
	updateInfoTable(info);

	Vector l;
	InfoEvent e = new InfoEvent(this, info);
    
	synchronized (this) { l = (Vector)simpListeners.clone(); }

	deliverInfoEvent(l, e);
    }

    private void deliverInfoEvent(final Vector l, final InfoEvent e) {
	if (SwingUtilities.isEventDispatchThread()) {
	    for (int i = 0; i < l.size(); i++) {
		((SimpListener)l.elementAt(i)).infoReceived(e);
	    }
	}
	else {
	    SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
			deliverInfoEvent(l, e);
		    }
		});
	}
    }

    private void noteLeaseEvent(IMA place, boolean create) {
	Vector l;
	LeaseEvent e = new LeaseEvent(this, place);
    
	synchronized (this) { l = (Vector)simpListeners.clone(); }

	deliverLeaseEvent(l, e, create);
    }

    private void deliverLeaseEvent(final Vector l, 
				   final LeaseEvent e,
				   final boolean create) {
	if (SwingUtilities.isEventDispatchThread()) {
	    for (int i = 0; i < l.size(); i++) {
		if (create)
		    ((SimpListener)l.elementAt(i)).subscriptionLeased(e);
		else
		    ((SimpListener)l.elementAt(i)).subscriptionLapsed(e);
	    }
	}
	else {
	    SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
			deliverLeaseEvent(l, e, create);
		    }
		});
	}
    }

    private void noteSubscriberEvent(IMA subscriber, boolean create) {
	Vector l;
	SubscriberEvent e = new SubscriberEvent(this, subscriber);
    
	synchronized (this) { l = (Vector)simpListeners.clone(); }

	deliverSubscriberEvent(l, e, create);
    }

    private void deliverSubscriberEvent(final Vector l, 
					final SubscriberEvent e,
					final boolean create) {
	if (SwingUtilities.isEventDispatchThread()) {
	    for (int i = 0; i < l.size(); i++) {
		if (create)
		    ((SimpListener)l.elementAt(i)).subscribed(e);
		else
		    ((SimpListener)l.elementAt(i)).unsubscribed(e);
	    }
	}
	else {
	    SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
			deliverSubscriberEvent(l, e, create);
		    }
		});
	}
    }

    private void noteMessageEvent(IMA to, IMA from, IMA replyTo, 
				  String type, String body) {
	Vector l;
	MessageEvent e = new MessageEvent(this, to, from, replyTo, type, body);
    
	synchronized (this) { l = (Vector)simpListeners.clone(); }

	deliverMessageEvent(l, e);
    }

    private void deliverMessageEvent(final Vector l, final MessageEvent e) {
	if (SwingUtilities.isEventDispatchThread()) {
	    for (int i = 0; i < l.size(); i++) {
		((SimpListener)l.elementAt(i)).messageReceived(e);
	    }
	}
	else {
	    SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
			deliverMessageEvent(l, e);
		    }
		});
	}
    }

    private void buddyInfoChanged(Info info) {
	if (debug)
	    log("Buddy info for " + infoString(info));
	Vector l;
	BuddyInfoEvent e = new BuddyInfoEvent(this, info);
    
	synchronized (this) { l = (Vector)simpListeners.clone(); }

	deliverBuddyInfoEvent(l, e);
    }

    private void deliverBuddyInfoEvent(final Vector l, 
				       final BuddyInfoEvent e) {
	if (SwingUtilities.isEventDispatchThread()) {
	    for (int i = 0; i < l.size(); i++) {
		((SimpListener)l.elementAt(i)).buddyInfoChanged(e);
	    }
	}
	else {
	    SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
			deliverBuddyInfoEvent(l, e);
		    }
		});
	}
    }    

    // Connection listener

    private Vector connectionListeners = new Vector();

    public synchronized void addConnectionListener(ConnectionListener el) {
	connectionListeners.add(el);
    }

    public synchronized void removeConnectionListener(ConnectionListener el) {
	connectionListeners.remove(el);
    }

    void closed(String msg) {
	noteConnectionEvent(msg, false);
    }

    private void noteConnectionEvent(String msg, boolean established) {
	Vector l;
	ConnectionEvent e = new ConnectionEvent(this, msg);
    
	synchronized (this) { l = (Vector)connectionListeners.clone(); }

	for (int i = 0; i < l.size(); i++) {
	    if (established)
		((ConnectionListener)l.elementAt(i)).connectionEstablished(e);
	    else
		((ConnectionListener)l.elementAt(i)).connectionClosed(e);
	}
    }

    // Profile listener

    private Vector profileListeners = new Vector();

    public synchronized void addProfileListener(ProfileListener el) {
	profileListeners.add(el);
    }

    public synchronized void removeProfileListener(ProfileListener el) {
	profileListeners.remove(el);
    }

    private void noteProfile(XmlProperties profile) {
	Vector l;
	ProfileChangedEvent e = new ProfileChangedEvent(this, profile);
    
	synchronized (this) { l = (Vector)profileListeners.clone(); }

	for (int i = 0; i < l.size(); i++) {
	    ((ProfileListener)l.elementAt(i)).profileChanged(e);
	}
    }

    // ACL listener

    private Vector aclListeners = new Vector();

    public synchronized void addACLListener(ACLListener el) {
	aclListeners.add(el);
    }

    public synchronized void removeACLListener(ACLListener el) {
	aclListeners.remove(el);
    }

    private void noteACL(ACL acl) {
	Vector l;
	ACLChangedEvent e = new ACLChangedEvent(this, acl);
    
	synchronized (this) { l = (Vector)aclListeners.clone(); }

	for (int i = 0; i < l.size(); i++) {
	    ((ACLListener)l.elementAt(i)).aclChanged(e);
	}
    }
}
