// $Id: ClientStart.java,v 1.14 2000/03/17 13:41:16 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.client.model;

import java.util.*;
import java.io.*;
import java.net.*;
import javax.swing.SwingUtilities;
import org.mitre.simp.util.*;
import org.mitre.simp.launcher.LoginResult;

/**
 * Implements the login protocol and then starts a client.
 * @author John D. Ramsdell
 * @version February 2000 
 */
public final class ClientStart
    implements XpListener
{
    private final IMA me;
    private final XpSocket xps;
    private final LoginResult result;
    private final ClientFactory cf;
    
    private ClientStart(IMA me, XpSocket xps,
			LoginResult result,
			ClientFactory cf) {
	this.me = me;
	this.xps = xps;
	this.result = result;
	this.cf = cf;
	xps.addXpListener(this);
    }

    // This method handles the second response of the login protocol.
    public void xpReceived(XpEvent e) {
	xps.removeXpListener(this);
	XmlProperties xp = e.getXp();
	String action = xp.remove(Protocol.ACTION);
	if (Protocol.SELF.equals(action)) {
	    result.loginSucceeded(null);
	    ClientModel cm = new ClientModel(me, xps, xp); 
	    cf.makeClient(cm);
	    cm.init();
	}
	else {
	    String message = xp.getProperty(Protocol.MESSAGE, "");
	    try {
		xps.close();
	    }
	    catch (IOException ex) { }
	    result.loginFailed(message);
	}
    }

    public void closed() {
	result.loginFailed("connection closed");
    }

    /**
     * This routine performs the login protocol, creates the client
     * given by the client factory, and then runs the communication
     * portion of a client.  It returns when the client can no longer
     * receive information from the server.  
     */
    public static void login(String user, char[] password, 
			     String host, LoginResult result,
			     ClientFactory cf) {
	// ****** HACK
	if (org.mitre.simp.launcher.SimpLauncher.getSimpProperties() == null) {
	    result.loginFailed("Update your launcher -- it's too old");
	    return;
	}
	// ******
	XpSocket xps;
	try {
	    IMA ima = new IMA(user, host);
	    XmlProperties req = Protocol.login(user);
	    XmlProperties resp = HostPost.post(host, req);
	    String action = resp.getProperty(Protocol.ACTION);
	    String nonce = resp.getProperty(Protocol.NONCE);
	    String opaque = resp.getProperty(Protocol.OPAQUE);
	    String alg = resp.getProperty(Protocol.ALGORITHM, "MD5");
	    host = resp.getProperty(Protocol.HOST, host);
	    String p = resp.getProperty(Protocol.PORT);
	    if (!Protocol.CHALLENGE.equals(action)
		|| nonce == null
		|| opaque == null
		|| alg == null
		|| p == null) {
		String msg = resp.getProperty(Protocol.MESSAGE, 
					      "login protocol error");
		result.loginFailed(msg);
		return;
	    }
	    int port = 0;
	    if (p != null) {
		try {
		    port = Integer.parseInt(p);
		}
		catch (NumberFormatException nfe) { }
	    }
	    if (port <= 0) {
		result.loginFailed("Bad port " + p);
		return;
	    }
	    Authorization authorization = null;
	    try {
		authorization = new Authorization(alg);
	    }
	    catch (java.security.NoSuchAlgorithmException nsae) {
		result.loginFailed("Digest algorithm " + alg + " missing");
		return;
	    }
	    String auth = null;
	    try {
		auth = authorization.getAuthorization(user, 
						      password,
						      nonce);
	    }
	    catch (UnsupportedEncodingException uee) {
		result.loginFailed("Unsupported encoding in login");
		return;
	    }
	    catch (IOException iox) {
		result.loginFailed("I/O exception in login");
		return;
	    }
	    Socket s = new Socket(host, port);
	    xps = new XpSocket(s);
	    // Build and send the second request of the login protocol.
	    xps.write(Protocol.connect(auth, opaque, 1, 0));
	    // The second response of the login protocol is handled
	    // by this object.
	    new ClientStart(ima, xps, result, cf);
	}
	catch (UnknownHostException ue) {
	    result.loginFailed("Host not known");
	    return;
	}
	catch (MalformedURLException ex) {
	    result.loginFailed("Address invalid");
	    return;
	}
	catch (ConnectException iox) {
	    result.loginFailed("Failed to connect to host");
	    return;
	}
	catch (NoRouteToHostException iox) {
	    result.loginFailed("Failed to find route to host");
	    return;
	}
	catch (Throwable t) {
	    result.loginFailed(t.getMessage());
	    t.printStackTrace();
	    return;
	}
	finally {
	    for (int i = 0; i < password.length; i++)
		password[i] = 0;
	}
	xps.run();
    }
}
