// $Id: MessageAppend.java,v 1.11 2000/04/07 19:28:34 gwilliam Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.client.released;

import java.awt.Color;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.text.html.*;
import java.awt.event.*;
import java.util.*;
import java.text.DateFormat;

import org.mitre.simp.util.IMA;

/**
 * A text pane for displaying conversation messages.  Allows the
 * messages to be displayed with optional sender identification and
 * time stamps.  BUG: the message in the pane are never purged so that
 * it is possible to run out of memory.
 * @version March 2000
 * @author Galen B. Williamson
 * @author John D. Ramsdell (adapted from TextAppend)
 */
public class MessageAppend extends JTextPane
{
    private final StyleContext sc = new StyleContext();
    private final Style normal = 
	sc.addStyle("normal", sc.getStyle(StyleContext.DEFAULT_STYLE));
    //  ss.addStyle("normal", ss.getStyle(StyleContext.DEFAULT_STYLE));    
    //private final StyleSheet ss = new StyleSheet();
    
    private final LinkedList messages = new LinkedList();

    private DefaultStyledDocument doc;
    //private HTMLDocument doc;
    private boolean showDate = false;
    private boolean longUser = false;

    private SimpleAttributeSet me_headerStyle = new SimpleAttributeSet();
    private SimpleAttributeSet you_headerStyle = new SimpleAttributeSet();

    private final Color meColor = new Color(255, 0, 0);
    private final Color youColor = new Color(0, 0, 255);

    private IMA me = null;

    /**
     * Create a MessageAppend text pane.
     */
    public MessageAppend() {
	this(null);
    }

    public MessageAppend(IMA me) {
	this.me = me;
	this.showDate = false;
	doc = createNewDocument();
	setStyledDocument(doc);
	setEditable(false);

	StyleConstants.setForeground(me_headerStyle, meColor);
	StyleConstants.setBold(me_headerStyle, false);
	StyleConstants.setFontFamily(me_headerStyle, 
				     StyleConstants.getFontFamily(normal));

	StyleConstants.setForeground(you_headerStyle, youColor);
	StyleConstants.setBold(you_headerStyle, false);
	StyleConstants.setFontFamily(you_headerStyle, 
				     StyleConstants.getFontFamily(normal));
    }

    /**
     * @return <code>false</code>
     */
    public boolean isFocusTraversable() {
	return false;
    }

    /**
     * Indicate whether messages should be shown with the full
     * Instant Messaging Address, or just the username, i.e.,
     * "foo" versus "foo@bar".
     */
    public void setShowLongUser(boolean show) {
	if (longUser != show) {
	    longUser = show;
	    refresh();
	}
    }

    /**
     * Indicate whether messages should be shown with time stamps.
     * Note: time stamps indicate the time the message was added,
     * since messages do not carry time stamps.
     */
    public void setShowDate(boolean show) {
	if (showDate != show) {
	    showDate = show;
	    refresh();
	}
    }

    /**
     * Add a message to the text pane with the indicated sender.
     */
    public void addMessage(IMA from, String body) {
	addMessage(new Message(from, body));
    }

    /**
     * Add a message to the text pane with no sender.
     */
    public void addMessage(String body) {
	addMessage(new Message(body));
    }

    /**
     * Add a message to the text pane with no sender.
     */
    public void addLine(String line) {
	addLines(new String[] { line });
    }

    /**
     * Add a message to the text pane with no sender, but follow each
     * element of the argument array with a newline.  
     */
    public void addLines(String[] lines) {
	addMessage(new Message(lines));
    }

    /**
     * Set the date format used to print messages' time stamps.
     */
    public void setDateFormat(DateFormat df) {
	this.df = df;
    }

    private DefaultStyledDocument createNewDocument() {
	return new DefaultStyledDocument(sc);
	// return new HTMLDocument(ss);
    }
    
    private void addMessage(Message m) {
	synchronized(messages) {
	    messages.addLast(m);
	}
	showMessage(m);
    }

    private void showMessage(final Message m) {
	if (SwingUtilities.isEventDispatchThread()) {
	    reallyShowMessage(m);
	}
	else {
	    SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
			reallyShowMessage(m);
		    }
		});
	}
    }

    private synchronized void reallyShowMessage(final Message m) {
	String header = m.makeHeader(longUser, showDate);
	String[] body = m.getBody();
	try {
	    doc.insertString(doc.getLength(), header, m.getHeaderStyle());
	} catch (BadLocationException ble1) { /* can't happen */ }
	for (int i = 0; i < body.length; i++) {
	    if (body[i] == null)
		body[i] = "";
	    try {
		doc.insertString(doc.getLength(), body[i], normal);
	    } catch (BadLocationException ble2) { /* can't happen */ }
	    doc.setLogicalStyle(doc.getLength() - 1, normal);
	    try {
		doc.insertString(doc.getLength(), "\n", null);
	    } catch (BadLocationException ble3) { /* can't happen */ }
	}
	if (getStyledDocument() == doc)
	    setCaretPosition(doc.getLength());
    }

    private synchronized void refresh() {
	new Thread() {
	    public void run () {
		StyledDocument oldDoc = doc;
		doc = createNewDocument();
		List l;
		synchronized(messages) {
		    l = (List)messages.clone();
		}
		Iterator i = l.listIterator();
		while (i.hasNext())
		    reallyShowMessage((Message)i.next());
		SwingUtilities.invokeLater(new Runnable() {
		    public void run() {
			setStyledDocument(doc);
			setCaretPosition(doc.getLength());
		    }});
	    }}.start();
    }

    private DateFormat df = DateFormat.getTimeInstance();

    private class Message {
	private IMA from;
	private String[] body;
	private Date date;
	private SimpleAttributeSet style;

	Message(String body) {
	    this(null, body);
	}
	Message(String[] body) {
	    this(null, body);
	}
	Message(IMA from, String body) {
	    this(from, new String[] { body });
	}
	Message(IMA from, String[] body) {
	    this.from = from;
	    this.body = body;
	    date = new Date(System.currentTimeMillis());
	    style = (from != null && !from.equals(me))
		? you_headerStyle
		: me_headerStyle;
	}
	String makeHeader(boolean longUser, boolean showDate) {
	    String s = "";
	    if (from != null)
		s += longUser ? from.toString() : from.getUser();
	    if (showDate) {
		if (from != null) s += " ";
		s += "["+df.format(date)+"]";
	    }
	    if (from != null || showDate) {
		s += ": ";
	    }
	    return s;
	}
	String[] getBody() { return body; }
	SimpleAttributeSet getHeaderStyle() { return style; }
    }
}
