// $Id: VertFillLayout.java,v 1.2 2000/02/21 12:45:49 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.client.released;

import java.awt.*;

/**
 * The Vertical Fill Layout class lays out items vertically and stretches 
 * them to fit the width.  The center component is stretched vertically
 * to fit the given space.
 * @version March 1997
 * @author John D. Ramsdell
 */
public class VertFillLayout
    implements LayoutManager, java.io.Serializable
{

    int hgap;			// Horizontal gap
    int vgap;			// Vertical gap
    Component centered;		// The compononent which is centered

    /**
     * Constructs a new Vertical Fill Layout.
     * @param hgap         horizontal gap
     * @param vgap         vertical gap
     */
    public VertFillLayout(int hgap, int vgap) {
	this.hgap = hgap;
	this.vgap = vgap;
    }

    /**
     * Constructs a new Vertical Fill Layout 
     * with default values for the 
     * horizontal and vertical gap.
     */
    public VertFillLayout() {
	this(5, 5);
    }

    /**
     * Adds the "Center" component to the layout.
     * This component is to determine the width of the layout.
     * @param name the name of the component
     * @param comp the the component to be added
     */
    public void addLayoutComponent(String name, Component comp) {
	if ("Center".equals(name))
	    centered = comp;
    }

    /**
     * Removes the "Center" component from the layout.
     * @param comp the component to remove
     */
    public void removeLayoutComponent(Component comp) {
	if (centered == comp)
	    centered = null;
    }

    /**
     * Returns the preferred dimensions for this layout given the components
     * in the specified target container.
     * @param target the component which needs to be laid out
     * @see Container
     * @see #minimumLayoutSize
     */
    public Dimension preferredLayoutSize(Container target) {
	Dimension dim = new Dimension(0, 0);
	int nmembers = target.getComponentCount();

	for (int i = 0 ; i < nmembers ; i++) {
	    Component m = target.getComponent(i);
	    if (m.isVisible()) {
		Dimension d = m.getPreferredSize();
		dim.width = Math.max(dim.width, d.width);
		if (i > 0) {
		    dim.height += vgap;
		}
		dim.height += d.height;
	    }
	}
	return addEdges(target, dim);
    }

    private Dimension addEdges(Container target, Dimension dim) {
	Insets insets = target.getInsets();
	dim.width += insets.left + insets.right + hgap*2;
	dim.height += insets.top + insets.bottom + vgap*2;
	return dim;
    }    

    /**
     * Returns the minimum dimensions needed to layout the components
     * contained in the specified target container.
     * @param target the component which needs to be laid out 
     * @see #preferredLayoutSize
     */
    public Dimension minimumLayoutSize(Container target) {
	Dimension dim = new Dimension(0, 0);
	int nmembers = target.getComponentCount();

	for (int i = 0 ; i < nmembers ; i++) {
	    Component m = target.getComponent(i);
	    if (m.isVisible()) {
		Dimension d = m.getMinimumSize();
		dim.width = Math.max(dim.width, d.width);
		if (i > 0) {
		    dim.height += vgap;
		}
		dim.height += d.height;
	    }
	}
	return addEdges(target, dim);
    }

    /**
     * Lays out the container.
     * @param target the specified component being laid out.
     * @see Container
     */
    public void layoutContainer(Container target) {
	Dimension targetSize = target.getSize();
	Dimension preferredSize = preferredLayoutSize(target);
	Insets insets = target.getInsets();
	int maxwidth = targetSize.width
	    - (insets.left + insets.right + hgap*2);
	int x = insets.left + hgap;
	int y = insets.top;
	int nmembers = target.getComponentCount();

	for (int i = 0; i < nmembers; i++) {
	    Component m = target.getComponent(i);
	    y += vgap;
	    if (m.isVisible()) {
		Dimension d = m.getPreferredSize();
		if (m == centered) {	// Stretch component to fill space.
		    int h 
			= targetSize.height - preferredSize.height + d.height;
		    if (h > 0)
			d.setSize(d.width, h);
		}
		m.setSize(maxwidth, d.height);
		m.setLocation(x, y);
		y += d.height;
	    }
	}
    }
    
    /**
     * Returns the String representation of this VertFillLayout's values.
     */
    public String toString() {
	return getClass().getName() + "[ hgap=" + hgap + ",vgap=" + vgap + "]";
    }
}
