// $Id: Login.java,v 1.2 2000/03/13 15:37:02 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.launcher;

import java.util.*;
import java.io.*;
import java.net.*;

/**
 * Login for the Simple Instant Messaging and Presence client.
 * The object downloads any needed files, creates the client object, 
 * and then runs the client.
 * @author John D. Ramsdell
 * @version February 2000 
 */
public class Login
    implements Runnable
{
    private final static int BUFSIZ = 256;
    private final String user;
    private final char[] password;
    private final String host;
    private final String font;
    private final String classpath;
    private final String classname;
    private final File cache;
    private final LoginResult lr;

    Login(String user, char[] password, String host, String font, 
	  String classpath, String classname, File cache, LoginResult lr) {
	this.user = user;
	this.password = password;
	this.host = host;
	this.font = font;
	this.classpath = classpath;
	this.classname = classname;
	this.cache = cache;
	this.lr = lr;
    }

    /**
     * The main actions are performed by this method.
     */
    public void run() {
	// Alert: add better parameter checking error messages!!!
	try {
	    URL[] urls = getURLs(classpath);
	    File[] files = getFiles(cache, urls);
	    FixedLengthDate fld = new FixedLengthDate();
	    for (int i = 0; i < urls.length; i++) {
		String protocol = urls[i].getProtocol();
		if (protocol != null && protocol.equals("http")) {
		    long mod = files[i].lastModified();
		    HttpURLConnection conn 
			= (HttpURLConnection)urls[i].openConnection();
		    conn.setRequestProperty("content-type",
					    "application/octet-stream");
		    if (mod > 0) {
			fld.setTime(new Date(mod));
			conn.setRequestProperty("If-Modified-Since", 
						fld.format());
		    }
		    conn.setRequestMethod("GET");
		    int rc = conn.getResponseCode();
		    if (rc == HttpURLConnection.HTTP_NOT_MODIFIED) {
			// The cache is good.
		    }
		    else if (rc == HttpURLConnection.HTTP_OK) {
			// Copy in new file
			copy(conn.getInputStream(), files[i]);
		    }
		    else if (mod == 0) {
			lr.loginFailed("Cannot get " + urls[i] + 
				       " " + conn.getResponseMessage());
			conn.disconnect();
			return;
		    }
		    conn.disconnect();
		    urls[i] = files[i].toURL();
		}
	    }
	    URLClassLoader cl = new URLClassLoader(urls);
	    Class c = cl.loadClass(classname);
	    Client client = (Client)c.newInstance();
	    client.init(user, password, host, font, lr);
	    client.run();
	}
	catch (MalformedURLException ex) {
	    lr.loginFailed("Bad URL in path");
	}
	catch (IOException ex) {
	    lr.loginFailed("I/O error updating path " + ex);
	}
	catch (ClassNotFoundException cx) {
	    lr.loginFailed("Cannot find class " + classname);
	}
	catch (IllegalAccessException ie) {
	    lr.loginFailed("Cannot access class " + classname);
	}
	catch (InstantiationException ie) {
	    lr.loginFailed("Cannot instantiate class " + classname);
	}
	catch (Throwable t) {
	    lr.loginFailed(t.getMessage());
	    t.printStackTrace();
	}
    }

    public static URL[] getURLs(String urlpath)
	throws MalformedURLException
    {
	if (urlpath == null)
	    return new URL[0];
	String[] urlstrs = string2array(urlpath, ';');
	URL[] urls = new URL[urlstrs.length];
	for (int i = 0; i < urlstrs.length; i++) 
	    urls[i] = new URL(urlstrs[i]);
	return urls;
    }

    private static File[] getFiles(File cache, URL[] url) {
	File[] file = new File[url.length];
	for (int i = 0; i < url.length; i++) {
	    String path = url[i].getFile();
	    int end = path.lastIndexOf('/');
	    String filename = end < 0 ? path : path.substring(end + 1);
	    file[i] = new File(cache, filename);
	}
	return file;
    }

    private void copy(InputStream in, File file)
	throws IOException
    {
	byte[] buf = new byte[BUFSIZ];
	FileOutputStream out = new FileOutputStream(file);
	for (;;) {
	    int n = in.read(buf);
	    if (n < 0)
		break;
	    out.write(buf, 0, n);
	}
	out.close();
	in.close();
    }

    /**
     * Break a string into an array of strings.
     * A sep character marks a breaking point.
     */
    public static String[] string2array(String str, char sep) {
	if (str == null)
	    return null;
	Vector v = new Vector();
	int s = 0;                    // s is the start index

	while (s < str.length()) {
	    if (str.charAt(s) == sep)
		s++;
	    else {
		int e = str.indexOf(sep, s); // e is the end index
		if (e < 0) {
		    v.addElement(str.substring(s));
		    break;
		}
		v.addElement(str.substring(s, e));
		s = e + 1;
	    }
	}

	String[] str_array = new String[v.size()];
	v.copyInto(str_array);
	return str_array;
    }
}
