// $Id: LoginPanel.java,v 1.1 2000/03/13 11:21:31 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.launcher;

import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import java.io.IOException;

/** 
 * Login panel.
 * @version July 1997
 * @author John D. Ramsdell
 * @author Sasha Porto Caskey
 */ 
public final class LoginPanel
    extends JPanel 
    implements ActionListener, FocusListener, LoginResult
{
    private final static int FIELD_SIZE = 20;
    private final JComboBox users = new JComboBox();
    private final JPasswordField password = new JPasswordField(FIELD_SIZE);
    private final Status status = new Status();
    private final static int PAD = 4;
    private SimpFrame sf;
    private SimpProperties props = null;

    /**
     * Constructs a new login panel.
     * @param name the initial value for the user field
     */
    public LoginPanel(SimpFrame sf, SimpProperties props) {

	this.props = props;
	this.sf = sf;

	Border inner = BorderFactory.createEmptyBorder(PAD, PAD, PAD, PAD);
	Border outer = BorderFactory.createEtchedBorder();
	setBorder(BorderFactory.createCompoundBorder(outer, inner));
	setLayout(new GridBagLayout());

	initComboBox();
	GridBagUtil gbu = new GridBagUtil(this,PAD);
	int row = 0;
	gbu.addLabel("User login:", row);
	gbu.addRightComponent(users, row++);
	gbu.addLabel("Password:", row);
	gbu.addRightComponent(password, row++);
	gbu.addLeftComponent(createButtonPanel(), row++);
	gbu.addStatus(status, row);
	password.addActionListener(this);
	users.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    password.requestFocus();
		}
	    });
	addFocusListener(this);
	status.showStatus("Ready");
    }

    void reloadProperties() {
	initComboBox();
    }

    /**
     * Save any modifications to property
     */
    void saveProperties() {
	updateUserList();

	try{
	    props.store();
	} catch (IOException iox) {
	    status.showStatus("Write to " + props.getFilename() + " failed");
	}
    }

    private JPanel createButtonPanel() {
	JPanel buttonPanel = new JPanel();
	JButton button;

	button = new JButton("Connect");
	button.setMnemonic('C');
	button.addActionListener(this);
	buttonPanel.add(button);
	sf.getRootPane().setDefaultButton(button);

	button = new JButton("Exit");
	button.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    sf.exit();
		}
	    });
	buttonPanel.add(button);
	return buttonPanel;
    }

    private void initComboBox() {
	String[] imas = props.getResourceTokenized(SimpLauncher.SIMP_USERS_KEY);	
	users.setEditable(true);
	users.removeAllItems();

	if (imas != null && imas.length > 0) {
	    for(int i=0; i < imas.length; i++) 
		users.insertItemAt(imas[i],i);
	    users.setSelectedIndex(0);
	}
    }

    private boolean ignoreLoginRequest = false;

    private synchronized void enableLoginRequest() {
	ignoreLoginRequest = false;
    }
   
    synchronized boolean ignoreLoginRequest() {
	if (ignoreLoginRequest)
	    return true;
	ignoreLoginRequest = true;
	return false;
    }

    /**
     * Reset panel in response to a failed login attempt.
     */
    public void loginFailed(final String message) {
	SwingUtilities.invokeLater(new Runnable() {
		public void run() {
		    loginInProcess = false;
		    password.setText("");
		    enableLoginRequest();
		    if (message == null)
			status.showStatus("Login failed");
		    else
			status.showStatus("Login failed: " + message);
		}
	    });
    }

    /* This should hide the frame on success. */
    public void loginSucceeded(final String message) {
	SwingUtilities.invokeLater(new Runnable() {
		public void run() {
		    loginInProcess = false;
		    password.setText("");
		    enableLoginRequest();
		    addNewUsers();
		    updateUserList();
		    if (message == null)
			status.showStatus("Login succeeded");
		    else
			status.showStatus("Login succeeded: " + message);
		}
	    });
	sf.close();
    }	

    /**
     * hand off focus to the correct field.
     */
    public void requestFocus() {
	if (users.getItemCount() > 0 && 
	    ((String)users.getItemAt(0)).trim().length() > 0 )
	    password.requestFocus();
	else
	    users.requestFocus();
    }

    /**
     * Handoff focus
     */
    public void focusGained(FocusEvent e) {
	requestFocus();
    }
  
    /**
     * This method does nothing.
     */
    public void focusLost(FocusEvent e) {}

    private void addNewUsers() {
	String ima = ((String)users.getSelectedItem()).trim();
	if(ima != null && ima.length() > 0) { 
	    int index = users.getSelectedIndex();
	    if (index < 0) {
		users.insertItemAt(ima,0);
	    } else if (index != 0) {
		users.removeItemAt(index);
		users.insertItemAt(ima,0);
		users.setSelectedIndex(0);
	    } 
	}
	saveProperties();
    }

    /**
     * Updates the JComboBox as well as the properties file.
     * It does <b>NOT</b>, however <tt>store()</tt> changes to the properties file.
     */ 
    private void updateUserList() {
	int n = users.getItemCount();
	if (n > 0) {
	    StringBuffer sb = new StringBuffer();
	    sb.append(users.getItemAt(0));
	    for(int i=1; i < n; i++) {
		sb.append(" ");
		sb.append(users.getItemAt(i));
	    }
	    props.setProperty(SimpLauncher.SIMP_USERS_KEY, sb.toString());
	}
    }
    
    /**
     * Tries to parse the IMA and return a hostname.
     *@return One of:
     *<ul>
     *<li>a hostname from the IMA.</li>
     *<li>the default hostname from the properties.</li>
     *<li>The empty String <tt>""</tt>.</li>
     *</ul>
     */
    protected String getHostFromIMA(String ima) {
	if (ima == null)
	    return "";

	int i = ima.indexOf('@');
	if (i < 0)
	    return props.getProperty(SimpLauncher.SIMP_HOST_KEY, "");

	return ima.substring(i+1);
    }

    /**
     * Parses the IMA and returns the user section. If no <tt>@</tt> is present returns the given IMA.
     */
    protected String getUserFromIMA(String ima) {
	if (ima == null)
	    return "";

	int i = ima.indexOf('@');
	if (i < 0)
	    return ima;
	return ima.substring(0,i);
    }

    private String searchHostnames() {
	int n = users.getItemCount();
	String h="";
	String ima="";

	if (n < 0) 
	    return "";
	
	for(int i=0; i < n; i++) {
	    ima = ((String)users.getItemAt(i)).trim();
	    h = getHostFromIMA(ima);
	    if (h.length() > 0) {
		props.setProperty(SimpLauncher.SIMP_HOST_KEY, h);
		return h;
	    }
	}
	return "";
    }
	
    /**
     * Action performed when text is entered
     */
    public void actionPerformed(ActionEvent e) {
	String ima = ((String)users.getSelectedItem()).trim();
	String u,h;

	h = getHostFromIMA(ima);
	u = getUserFromIMA(ima);

	if (u.length() == 0) 
	    return;

	if (h.length() == 0) 
	    h = searchHostnames();
	    
	char[] p = password.getPassword();
	if (h.equals("")) {
	    status.showStatus("Login with usename@host");
	    JOptionPane.showMessageDialog(null, "There is no default hostname defined.\nEither set one using the properties editor or\nuse the username@hostname notation as a login\n(e.g. dude@simpserver.mitre.org)", "Oops! No default hostname defined", 
					  JOptionPane.ERROR_MESSAGE);
	} else {
	    status.showStatus("Sending login information");
	    notifyOfLogin(u, p, h);
	}
    }

    private Vector loginListeners = new Vector();

    /**
     * Adds a login listener
     */
    public synchronized void addLoginListener(LoginListener ll) {
	loginListeners.addElement(ll);
    }

    /**
     * Removes a login listener
     */
    public synchronized void removeLoginListener(LoginListener ll) {
	loginListeners.removeElement(ll);
    }
  
    private volatile boolean loginInProcess = false;
    /**
     * Notify login listeners of the receipt of a login
     */
    private void notifyOfLogin(String user, char[] password, String host) {
	Vector l;
	
	if(!loginInProcess) {
	    LoginEvent e = new LoginEvent(this, user, password, host);
	
	    synchronized (this) { 
		l = (Vector)loginListeners.clone(); 
		loginInProcess = true;
	    }

	    for (int i = 0; i < l.size(); i++) {
		((LoginListener)l.elementAt(i)).performLogin(e);
	    }
	}
    }
}
