// $Id: SimpProperties.java,v 1.1 2000/03/13 11:21:32 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.launcher;

import java.util.*;
import java.io.*;
import java.net.URL;
import java.net.MalformedURLException;

/**
 * Starter application for running a Simp Client.
 * @version Feburary 2000
 * @author Sasha Porto Caskey
 */ 

public class SimpProperties 
{
    private String filename=null;
    private String header=null;
    private boolean changed = false;
    private Properties props = new Properties();

    /**
     * SimpProperties class
     * @param filename Properties filename.
     * @param header Properties header string.
     */

    public SimpProperties(String filename, String header) 
	throws IOException {
	this.filename = filename;
	this.header = header;
	FileInputStream fin = new FileInputStream(filename);
	BufferedInputStream bin = new BufferedInputStream(fin);
	props.load(bin);
	bin.close();
    }

    /**
     * SimpProperties class. Uses <tt>SIMP Properties</tt> as default header.
     * @param filename Properties filename.
     */
    public SimpProperties(String filename) 
	throws IOException {
	this(filename,"SIMP Properties");
    }

    private SimpProperties() {
    }

    /**
     * Clones this object
     */
    public Object clone() {
	SimpProperties sp = new SimpProperties();
	sp.props = (Properties)this.props.clone();
	sp.filename = filename;
	sp.header = header;
	sp.changed = this.changed;
	return sp;
    }
	
    /**
     * Uses the default filename and header if they are set
     */
    public void store() 
	throws IOException {
	store(header);
    }

    /**
     * Uses the default filename if it is set.
     * @param header Header string for properties file
     */
    public void store(String header) 
	throws IOException {
	if(changed) {
	    FileOutputStream fos = new FileOutputStream(filename);
	    BufferedOutputStream bos = new BufferedOutputStream(fos);
	    props.store(bos,header);
	    bos.close();
	    changed = false;
	}
    }

    /**
     * Retrieve current filename associated with this property file
     * @return filename String representation of filename.
     */
    public String getFilename() {
	return filename;
    }

    /**
     * Maps to a java.util.Hashtable.containsKey()
     * @return true if the key exists
     */
    public boolean containsKey(String key) {
	return props.containsKey(key);
    }

    /**
     * Maps to a java.util.Properties.setProperty()
     * @return String instead of an Object
     */
    public String setProperty(String key, String value) {
	changed = true;
	return (String)props.setProperty(key,value);
    }

    /**
     * Maps to a java.util.Properties.remove()
     * @return String instead of an Object
     */
    public String remove(String key) {
	changed = true;
	return (String)props.remove(key);
    }

    /**
     * Maps to a java.util.Properties.getProperty()
     */
    public String getProperty(String key) {
	return props.getProperty(key);
    }

    /**
     * Maps to a java.util.Properties.getProperty()
     */
    public String getProperty(String key, String defaultValue) {
	return props.getProperty(key,defaultValue);
    }

    /**
     * Get a URL from a ResourceBundle.
     * @param nm name of key
     * @return resource as a URL or null on error
     */
    public URL getResourceURL(String nm) {
	String s = getResourceString(nm);
	if (s == null) return null;
	try {
	    return new URL(s);
	} catch(MalformedURLException mue) {
	    return null;
	}
    }

    /**
     * Get a string from a ResourceBundle.
     * @param nm name of key
     * @return resource as a string or null on error
     */
    public String getResourceString(String nm) {
	return nm == null ? null : props.getProperty(nm);
    }

    /**
     * Get a string as an array of tokens from a ResourceBundle.
     * @param nm name of key
     * @return resource as a string or null on error
     */
    public String[] getResourceTokenized(String nm) {
	return getResourceTokenized(nm,null);
    }

    /**
     * Get a string as an array of tokens from a ResourceBundle.
     * @param nm name of key
     * @param delim delimiter string to pass on to <code>StringTokenizer</code>
     * @return resource as a string or null on error
     */
    public String[] getResourceTokenized(String nm, String delim) {
	String tokens = getResourceString(nm);
	if (tokens == null) return null;
	StringTokenizer t;
	if (delim == null) {
	    t = new StringTokenizer(tokens);
	} else {
	    t = new StringTokenizer(tokens, delim);
	}
	Vector v = new Vector();
	while (t.hasMoreTokens())
	v.addElement(t.nextToken());
	String[] tokenized = new String[v.size()];
	return (String[])v.toArray(tokenized);
    }

    /**
     * Get an integer from a ResourceBundle.
     * @param nm name of key
     * @param defaultValue value on error
     * @return resource as an integer or default value on error
     */
    public int getResourceInt(String nm, int defaultValue) {
	String str = getResourceString(nm);
	if (str == null)
	    return defaultValue;
	try {
	    return Integer.parseInt(str);
	}
	catch (NumberFormatException ex) {
	    return defaultValue;
	}
    }

    /**
     * Get a boolean from a ResourceBundle.
     * @param nm name of key
     * @return resource as a boolean or the false value on error
     */
    public boolean getResourceBoolean(String nm) {
	String str = getResourceString(nm);
	return Boolean.valueOf(str).booleanValue();
    }
}
