// $Id: ContactPlace.java,v 1.16 2000/04/11 16:56:55 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.server;

import java.util.*;
import java.io.*;
import java.net.*;
import org.mitre.simp.util.*;

/**
 * The place associated with a user.  A contact place provides
 * presence information, and when the user is connected, a relay for
 * instant messages.  As an XML Service, it responds to requests from
 * a notification connection.
 * @author John D. Ramsdell
 * @version February 2000 
 */
final class ContactPlace
    extends XmlService
{
    // maximum duration of a subscription
    private final static long MAX_DURATION = 25 * 60 * 60 * 1000;
    // constants used during file name generation.
    private final static String profExt = "_profile.xml";
    private final static String aclExt = "_acl.xml";

    private Endpoint endpoint;	// connection point of an online user
    private DateTime online;	// null when user not online
    private volatile String message; // cached message
    private final List subscriptions = new LinkedList();

    private final String user;
    private final SimpService service;
    private final Logger logger;
    private final Profile profile;
    private final Judge judge;

    /**
     * Creates the place associated with a user.
     */
    ContactPlace(String user, SimpService service) {
	this.user = user;
	this.service = service;
	logger = service.getLogger();
	profile = new Profile(user + profExt, 
			      Protocol.SERVER_USER + profExt);
	judge = new Judge(user + aclExt,
			  Protocol.SERVER_USER + aclExt);
    }

    void init() {
	super.init();
	getACL();		// load ACL from the file system
	getProfile();		// load profile from the file system
    }

    // some getters and setters

    String getUser() {
	return user;
    }

    SimpService getSimpService() {
	return service;
    }

    Logger getLogger() {
	return logger;
    }

    private void log(String msg, Throwable t) {
	logger.log(user + "'s place: " + msg, t);
    }

    XmlProperties getACL() {
	try {
	    return judge.getACL().getXp();
	}
	catch (IOException iox) {
	    return null;
	}
    }

    XmlProperties getProfile() {
	try {
	    XmlProperties prof = profile.getProfile();
	    message = prof.getProperty(Protocol.MESSAGE, "");
	    return prof;
	}
	catch (IOException iox) {
	    message = "";
	    return null;
	}
    }

    void setProfile(XmlProperties prof) 
	throws IOException
    {
	profile.setProfile(prof);
	String msg = prof.getProperty(Protocol.MESSAGE, "");
	if (message == null || !message.equals(msg)) {
	    message = msg;
	    notifyOfChange();
	}
    }

    void setACL(XmlProperties acl) 
	throws IOException
    {
	judge.setACL(acl);
    }

    // starts an online session with this user
    synchronized void login(Endpoint ep) {
	Endpoint old = endpoint;
	if (old != null)	// kill old connection if it exists
	    try {
		old.getXpSocket().kill();
	    }
	    catch (IOException iox) { }
	endpoint = ep;
	online = new DateTime();
	notifyOfChange();	// tell subscribers about this
    }

    void noteSubscribers() {	// tell user of the current 
	Endpoint ep = endpoint;	// set of subscribers.
	if (ep == null)		// used during login.
	    return;
	Subscription[] a = getSubscriptions();
	for (int i = 0; i < a.length; i++) {
	    IMA subscriber = a[i].getSubscriber();
	    XmlProperties command = Protocol.noteSubscription(subscriber);
	    try {
		ep.write(command);
	    }
	    catch (IOException iox) {
		log("write of subscriptions failed", iox);
	    }
	}
    }

    synchronized void logout() { // terminate an online connection
	endpoint = null;
	online = null;
	log("user logged out", null);
	notifyOfChange();	// tell subscribers about this
    }

    synchronized void destroy() { // destroy this place.
	Endpoint ep = endpoint;	// used during an orderly shutdown.
	if (ep != null)
	    try {
		ep.close();
	    }
	    catch (IOException iox) { }
    }

    private void notifyOfChange() { // send note change messages
	if (message == null)
	    getProfile();
	Subscription[] a = getSubscriptions();
	if (a.length == 0)
	    return;
	try {
	    IMA regarding = new IMA(user);
	    IMA me = service.getMe();
	    XmlProperties xp 
		= Protocol.noteChange(me, regarding, online, message);
	    Pool p = service.getPool();
	    if (!p.fork(changeTask(me, regarding, a, xp))) {
		log("cannot create a note change thread", null);
	    }
	}
	catch (UnknownHostException ue) {
	    log("cannot make subject in notifyOfChange", ue);
	}
    }

    private Subscription[] getSubscriptions() {
	synchronized (subscriptions) {
	    List l = subscriptions;
	    Subscription[] a = new Subscription[l.size()];
	    return (Subscription[])l.toArray(a);
	}
    }

    private Runnable changeTask(final IMA self, final IMA regarding,
				final Subscription[] l, 
				final XmlProperties req) {
	return new Runnable() {
		public void run() {
		    for (int i = 0; i < l.length; i++) {
			Subscription s = l[i];
			IMA subscriber = s.getSubscriber();
			req.setProperty(Protocol.TO, subscriber.toString());
			XmlProperties resp = service.deliver(subscriber, 
							     self, req);
			String action = resp.getProperty(Protocol.ACTION);
			if (!Protocol.OKAY.equals(action)) 
			    if (removeSubscription(subscriber, null))
				noteSubscriptionLapse(subscriber);
		    }			    
		}
	    };
    }

    // tell user of a new subscriber
    void noteSubscription(IMA subscriber) {
	Endpoint ep = endpoint;
	if (ep != null) {
	    XmlProperties xp = Protocol.noteSubscription(subscriber);
	    try {
		ep.write(xp);
	    }
	    catch (IOException iox) {
		log("write of note subscription failed", iox);
	    }
	}
    }

    // tell user a subscription lapsed
    void noteSubscriptionLapse(IMA subscriber) {
	Endpoint ep = endpoint;
	if (ep != null) {
	    XmlProperties xp = Protocol.noteSubscriptionLapse(subscriber);
	    try {
		ep.write(xp);
	    }
	    catch (IOException iox) {
		log("write of note subscription lapse failed", iox);
	    }
	}
    }

    // notify subscriber of premature subscription termination
    XmlProperties noteSubscriberEnd(IMA subscriber) {
	IMA to = subscriber;
	IMA from = service.getMe();
	IMA regarding;
	try {
	    regarding = new IMA(user);
	}
	catch (UnknownHostException uhe) {
	    return Protocol.error("Cannot create user IMA");
	}
	XmlProperties req 
	    = Protocol.noteSubscriptionEnd(to, from, regarding,
					   online, message);
	return service.deliver(to, from, req);
    }

    long addSubscription(String opaque, IMA subscriber, long duration) {
	if (duration > MAX_DURATION || duration < 0)
	    duration = MAX_DURATION;
	Subscription lease = new Subscription(this, opaque,
					      subscriber, duration);
	synchronized (subscriptions) {
	    List l = subscriptions;
	    for (ListIterator i = l.listIterator(); i.hasNext();) {
		Subscription s = (Subscription)i.next();
		if (subscriber.equals(s.getSubscriber())) {
		    i.remove();
		    s.setAlive(false);
		}
	    }
	    l.add(lease);
	}
	service.addSubscription(lease);
	return duration;
    }

    // opaque = null to override opaque check.
    boolean removeSubscription(IMA subscriber, String opaque) {
	synchronized (subscriptions) {
	    List l = subscriptions;
	    for (ListIterator i = l.listIterator(); i.hasNext();) {
		Subscription s = (Subscription)i.next();
		if (subscriber.equals(s.getSubscriber())
		    && (opaque == null || opaque.equals(s.getOpaque()))) {
		    i.remove();
		    s.setAlive(false);
		    return true;
		}
	    }
	}
	return false;
    }

    // local message relays
    XmlProperties sendLocal(int op, IMA to, IMA from, XmlProperties req) {
	if (!judge.permit(from, op, false))
	    return Protocol.error("Permission denied");
	Endpoint ep = endpoint;
	if (ep == null)
	    return Protocol.error("User not found");
	try {
	    return ep.post(req);
	}
	catch (IOException ex) {
	    try {
		ep.close();
	    }
	    catch (IOException iox) {
	    }
	    log("local message could not be delivered", ex);
	    return Protocol.error("Local message could not be delivered");
	}
    }

    XmlProperties send(IMA to, IMA from, XmlProperties req) {
	return sendLocal(ACL.OP_SEND, to, from, req);
    }

    XmlProperties noteChange(IMA to, IMA from, XmlProperties req) {
	return sendLocal(ACL.OP_CHANGE, to, from, req);
    }

    XmlProperties noteSubscriptionEnd(IMA to, IMA from, XmlProperties req) {
	return sendLocal(ACL.OP_END, to, from, req);
    }

    // get presence information
    XmlProperties fetch(IMA to, IMA from, XmlProperties req) {
	if (!judge.permit(from, ACL.OP_FETCH, false))
	    return Protocol.error("Permission denied");
	if (message == null)
	    getProfile();	// make sure profile is loaded
	return Protocol.describe(to, online, message);
    }

    XmlProperties subscribe(IMA to, IMA from, XmlProperties req) {
	if (!judge.permit(from, ACL.OP_SUBSCRIBE, false))
	    return Protocol.error("Permission denied");
	if (message == null)
	    getProfile();	// make sure profile is loaded
	if (to.equals(from)) {
	    Protocol.error("Cannot subscribe to oneself");
	}
	long duration = req.getPropertyAsLong(Protocol.DURATION, 0);
	if (duration == 0)
	    return Protocol.error("Missing or bad duration");
	String opaque = service.nextOpaque();
	duration = addSubscription(opaque, from, duration);
	noteSubscription(from);	// tell user
	return Protocol.describe(to, online, message, duration, opaque);
    }

    XmlProperties unsubscribe(IMA to, IMA from, XmlProperties req) {
	if (!judge.permit(from, ACL.OP_UNSUBSCRIBE, false))
	    return Protocol.error("Permission denied");
	if (message == null)
	    getProfile();	// make sure profile is loaded
	String opaque = req.getProperty(Protocol.OPAQUE, "");
	if (removeSubscription(from, opaque)) {
	    noteSubscriptionLapse(from); // tell user
	    return Protocol.okay("Unsubscribed");
	}
	else {
	    return Protocol.error("Unsubscribe request failed");
	}
    }

    // This routine handles requests made by the user's client
    // through the endpoint.
    XmlProperties service(XmlProperties req) {
	String action = req.getProperty(Protocol.ACTION);
	if (action == null) 
	    return service.missingAction(req);
	else if (Protocol.LOGIN.equals(action)) 
	    return Protocol.error("Login not allowed");
	else if (Protocol.GET_ACL.equals(action)) {
	    XmlProperties acl = getACL();
	    if (acl == null)
		return Protocol.error("Cannot get ACL");
	    XmlProperties resp = (XmlProperties)acl.clone();
	    resp.setProperty(Protocol.ACTION, Protocol.SELF);
	    return resp;
	}
	else if (Protocol.GET_PROFILE.equals(action)) {
	    XmlProperties pref = getProfile();
	    if (pref == null)
		return Protocol.error("Cannot get profile");
	    XmlProperties resp = (XmlProperties)pref.clone();
	    resp.setProperty(Protocol.ACTION, Protocol.SELF);
	    return resp;
	}
	else if (Protocol.SET_ACL.equals(action)) {
	    req.remove(Protocol.ACTION);
	    try {
		setACL(req);
		return Protocol.okay("Done");
	    }
	    catch (IOException iox) {
		return Protocol.error("Store failed: "+iox);
	    }
	}
	else if (Protocol.SET_PROFILE.equals(action)) {
	    req.remove(Protocol.ACTION);
	    try {
		setProfile(req);
		return Protocol.okay("Done");
	    }
	    catch (IOException iox) {
		return Protocol.error("Store failed");
	    }
	}
	else if (Protocol.DROP_SUBSCRIPTION.equals(action)) {
	    IMA subscriber;
	    try {
		subscriber = new IMA(req.getProperty(Protocol.SUBSCRIBER));
	    }
	    catch (UnknownHostException ue) {
		return Protocol.error("Subscriber field not an address");
	    }
	    if (removeSubscription(subscriber, null)) {
		noteSubscriptionLapse(subscriber);
		noteSubscriberEnd(subscriber);
		return Protocol.okay("Subscriber removed");
	    }
	    else {
		return Protocol.error("Subscriber not found");
	    }
	}
	else {
	    // forward all remaining requests to the main dispatch loop
	    IMA to;
	    try {
		to = new IMA(req.getProperty(Protocol.TO));
	    }
	    catch (UnknownHostException ue) {
		return Protocol.error("To field not an address");
	    }
	    IMA from;
	    try {
		from = new IMA(req.getProperty(Protocol.FROM));
	    }
	    catch (UnknownHostException ue) {
		return Protocol.error("From field not an address");
	    }
	    return service.deliver(to, from, req);
	}
    }
}
