// $Id: Pool.java,v 1.11 2000/04/13 13:09:50 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.server;

/**
 * A pool of threads.
 * @author John D. Ramsdell
 * @version February 2000 
 */
public final class Pool
{
    private static final int CACHE_SIZE = 128;
    private final int limit;
    private final Logger logger;
    private int count = 0;	// Number of threads in use.
    private final WorkerCache cache = new WorkerCache(CACHE_SIZE);
    private volatile boolean poisoned = false;

    /**
     * Creates a pool of threads for running tasks.
     * @param limit when nonzero, limit is maximum number of threads created
     * @param logger for error reports
     */
    public Pool(int limit, Logger logger) {
	this.limit = limit;
	this.logger = logger;
    }

    /**
     * Run a task on a thread maintained by this pool.
     * @param task to be run by a thread in the pool
     * @return false if no thread is available due to the thread limit
     */
    public boolean fork(Runnable task) {
	if (task == null)
	    return true;
	if (poisoned)
	    return false;
	if (limit > 0) 
	    synchronized (this) {
		if (count >= limit)
		    return false;	// thread limit reached
		count++;
	    }
	Worker bee = cache.get();
	if (bee != null)
	    bee.setTask(task);
	else
	    new Worker(task).start();
	return true;
    }
       
    /**
     * Release the resources associated with this pool.
     */
    public void destroy()
    {
	poisoned = true;	// This is the only place this variable is set
	cache.destroy();
    }

    /**
     * Place an idle worker in the cache.
     */
    void idle(Worker bee) {
	if (poisoned) 
	    bee.kill();
	if (limit > 0)
	    synchronized (this) {
		count--;
	    }
	cache.put(bee);
    }

    void log(String message, Throwable t) {
	logger.log(message, t);
    }

    /**
     * A cache of workers.
     */
    private static class WorkerCache 
    {
	private final int limit;
	private final Worker[] workers;
	private int count = 0;	// Entries in the queue
	private int top = 0;	// Entry place in the queue

	WorkerCache(int limit) {
	    this.limit = limit;
	    workers = new Worker[limit];
	}

	private int incr(int i) {
	    return (i + 1) % limit;
	}

	private int diff(int i, int j) {
	    int k = i - j;
	    return k >= 0 ? k : k + limit;
	}
		

	/**
	 * Add a worker to the cache.
	 */
	synchronized void put(Worker bee) {
	    Worker w = workers[top];
	    workers[top] = bee;
	    top = incr(top);
	    if (w == null)
		count++;
	    else		// kill old worker
		w.kill();
	}
	
	/**
	 * Get a live worker from the cache or return null.
	 */
	synchronized Worker get() {
	    for (int bottom = diff(top, count);
		 count > 0;
		 bottom = incr(bottom)) {
		Worker bee = workers[bottom];
		workers[bottom] = null;
		count--;
		if (bee != null) {
		    if (bee.isAlive())
			return bee;
		    else
			bee.kill();
		}
	    }
	    return null;
	}
	
	/**
	 * Kill cached workers.
	 */
	synchronized void destroy() {
	    for (int bottom = diff(top, count);
		 count > 0;
		 bottom = incr(bottom)) {
		Worker bee = workers[bottom];
		workers[bottom] = null;
		count--;
		if (bee != null)
		    bee.kill();
	    }
	}
    }

    /**
     * A worker
     */
    private class Worker
	extends Thread
    {
	private Runnable task;
	private boolean poisoned;

	Worker(Runnable task) {
	    this.task = task;
	}

	/**
	 * Set the task.
	 * @param task the task
	 */
	synchronized void setTask(Runnable task) {
	    this.task = task;
	    notifyAll();
	}

	/**
	 * Destroy worker
	 */
	synchronized void kill() {
	    poisoned = true;
	    notifyAll();
	}

	/**
	 * Get next task
	 */
	private synchronized Runnable getNextTask() {
	    while (!poisoned && task == null) {
		try {		// wait for notify from
		    wait(0);	// setTask or kill
		}
		catch (InterruptedException ex) {
		}
	    }
	    Runnable t = task;
	    task = null;
	    return t;
	}

	public void run() {
	    for (;;) {
		Runnable t = getNextTask();
		if (poisoned)
		    return;
		try {
		    t.run(); // Run task
		}
		catch (Throwable thr) {
		    log("pool: thread exception " + thr, thr);
		}
		finally {
		    idle(this);
		}
	    }
	}
    }
}
