// $Id: Queue.java,v 1.7 2000/04/11 16:56:55 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.server;

import java.util.*;
import java.io.*;
import java.net.*;
import org.mitre.simp.util.*;

/**
 * Subscription queue.  An instance of this class is used to flush
 * lapsed subscrptions.
 * @author John D. Ramsdell
 * @version February 2000 
 */
final class Queue
    implements Runnable
{
    private final List l = new LinkedList();
    private Thread t;		// thread used to reap subscriptions

    /**
     * The subscriptions are stored as a sorted linked list.  The list
     * is sorted based on the time at which the subscription lapses.
     */
    synchronized void add(Subscription lease) {
	ListIterator i = l.listIterator();
	while (i.hasNext()) {
	    Subscription s = (Subscription)i.next();
	    if (s.getDone() >= lease.getDone()) {
		i.previous();
		break;
	    }
	}
	i.add(lease);
	if (t == null || !t.isAlive()) {
	    t = new Thread(this);
	    t.setDaemon(true);
	    t.start();		// start subscription reaper
	}
	else
	    notify();		// restart subscription reaper
    }

    public void run() {
	for (;;) {
	    Subscription s = getLapsedSubscription();
	    if (s == null)	// get next subscription
		return;
	    s.noteSubscriptionLapse(); // and finalize it
	}
    }

    private synchronized Subscription getLapsedSubscription() {
	try {
	    for (;;) {
		long now = System.currentTimeMillis();
		for (;;) {
		    if (l.isEmpty()) // exit reaper
			return null;
		    long t = ((Subscription)l.get(0)).getDone();
		    if (t > now) {
			wait(t - now); // wait for lapse
			break;
		    }
		    Subscription s = (Subscription)l.remove(0);
		    if (s.isAlive()) // ignore dead subscriptions
			return s;
		    
		}
	    }
	}
	catch (InterruptedException iex) {
	    return null;	// exit reaper
	}
    }
}
