// $Id: SimpServer.java,v 1.9 2000/04/11 16:56:56 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.server;

import java.io.*;
import java.net.ServerSocket;
import org.mitre.simp.util.*;

/**
 * Main entry point for running a Simp Server.
 * @version February 2000
 * @author John D. Ramsdell
 */ 

class SimpServer
{
    private SimpServer() { }

    private final static String SIMP_PROPERTIES = "simpserv.properties";
    private final static int SIMP_MINOR_VERSION = 0;
    private final static int SIMP_MAJOR_VERSION = 1;

    /**
     * Starts a SimpServer.
     * If present first arg is the property file.
     */
    public static void main(String args[]) {
	if (args.length > 1) {
	    usage();
	    return;
	}

	// Load the properties file.
	SimpProperties props = new SimpProperties();
	String filename = args.length == 0 ? SIMP_PROPERTIES : args[0];
	loadProperties(props, filename);

	// Create a logger
	Logger logger = new Logger(props);

	// Create a pool of threads for the router service
	String key = SimpProperties.ROUTER_POOL_LIMIT_KEY;
	int limit = props.getPropertyAsInt(key, 0);
	logger.log("server: router pool limit is " + limit);
	Pool pool = new Pool(limit, logger);

	// Create the server socket for the routing connection
	// Shut down server by closing this socket
	ServerSocket ss = makeRouterSocket();

	// Create the SIMP service
	SimpService serv = new SimpService(props, logger, ss);

	serv.init();		// Start the notification connection
	ServiceRunner sr = new ServiceRunner(serv, pool, ss, logger);
	System.err.println("SIMP Server running");
	sr.run();		// Start the routing connection
	serv.destroy();		// Close down nicely if possible
	pool.destroy();
    }

    private static void usage() {
	System.err.println("Usage: java " + SimpServer.class.getName()
			   + " [ PROPERTY FILE ]"); 
	System.exit(1);
    }

    private static void loadProperties(SimpProperties props, String filename) {
	String message = null;
	try {
	    FileInputStream fin = new FileInputStream(filename);
	    BufferedInputStream bin = new BufferedInputStream(fin);
	    props.load(bin);
	    bin.close();
	}
	catch (FileNotFoundException fnfe) {
	    message = "Properties file " + filename + " not found";
	}
	catch (IOException ioe) {
	    message = "An error occured while reading file " + filename;
	}
	if (message != null) {
	    System.err.println(message);
	    System.exit(1);
	}
    }

    private static ServerSocket makeRouterSocket() {
	try {
	    return new ServerSocket(HostPost.SIMP_PORT);
	}
	catch (IOException iox) {
	    System.err.println("Failed to create a SIMP socket at port " + 
			       HostPost.SIMP_PORT);
	    System.exit(1);
	    return null;
	}
    }

    public static int getMajorVersion() {
	return SIMP_MAJOR_VERSION;
    }

    public static int getMinorVersion() {
	return SIMP_MINOR_VERSION;
    }
}
