// $Id: SimpService.java,v 1.17 2000/04/11 16:56:56 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.server;

import java.util.*;
import java.io.*;
import java.net.*;
import org.mitre.simp.util.*;

/**
 * An instance of this class provides routing service as an XML
 * Service.  As part of its initialization, notification service is
 * started.
 * @author John D. Ramsdell
 * @version February 2000
 */
final class SimpService
    extends XmlService
{
    private ServerSocket notifierSocket; // close to stop notification
    private final ServerSocket routerSocket; // close to stop routing
    private int port;		// port for notification
    private final Nonce nonceGen = new Nonce();
    private final static String DEFAULT_PASSWORDS = "passwords.properties";
    private Password pw;	// passwords for users
    private final static String DEFAULT_ALGORITHM = "MD5";
    private Authorization authorization;
    private IMA me;		// the server's IMA (notifier@HOST)
    private final static long DEFAULT_COOKIE_TIME_OUT = 10 * 1000;
    private long cookie_time_out; // time out in login protocol
    private final static int SO_TIMEOUT = 4 * 60 * 1000;
    private Pool notifierPool;	// thread pool for the notifier

    // Verbose mode flag.
    private boolean verbose;

    // maps opaque strings to cookies
    private final CookieTable cookieTable = new CookieTable();

    // maps a user name to a contact place
    private final ContactPlaceTable cpt = new ContactPlaceTable(this);

    // a queue of subscriptions used to reap lapsed subscriptions
    private final Queue q = new Queue();

    private final SimpProperties props;
    private final Logger logger;

    /**
     * Create an XML service for routing.
     */
    SimpService(SimpProperties props, Logger logger, 
		ServerSocket routerSocket) {
	this.props = props;
	this.logger = logger;
	this.routerSocket = routerSocket;
    }

    boolean isVerbose() {
	return verbose;
    }

    void setVerbose(boolean verbose) {
	this.verbose = verbose;
    }

    Pool getPool() {
	return notifierPool;
    }

    Logger getLogger() {
	return logger;
    }

    IMA getMe() {
	return me;
    }

    void addSubscription(Subscription s) {
	q.add(s);
    }

    String nextOpaque() {
	return nonceGen.nextNonce();
    }

    // used by the contact place table
    boolean hasAccount(String user) {
	try {
	    return pw.getPassword(user) != null;
	}
	catch (IOException iox) {
	    return false;
	}
	
    }

    /* The init process should look for a file encoding the 
       subscription queue and send out subscription
       end notifications for each item on the queue.

       The destroy process should save the subscription queue.

       Periodically, the server should save the state of the
       subscription queue in case the server crashes.
    */

    /** 
     * Initialze the router service
     */
    void init() {
	super.init();
	// set parameters from properties
	// set verbose mode
	String key = SimpProperties.VERBOSE_KEY;
	verbose = props.getPropertyAsBoolean(key);

	// set cookie time out used in login
	key = SimpProperties.COOKIE_TIME_OUT_KEY;
	long default_timeout = DEFAULT_COOKIE_TIME_OUT;
	cookie_time_out = props.getPropertyAsLong(key, default_timeout);
	log("init: cookie time out is " + cookie_time_out);

	key = SimpProperties.PASSWORD_FILE_NAME_KEY;
	String passwords = props.getProperty(key, DEFAULT_PASSWORDS);
	log("init: password file is " + passwords);
	pw = new Password(new File(passwords));

	key = SimpProperties.DIGEST_ALGORITHM_KEY;
	String algorithm = props.getProperty(key, DEFAULT_ALGORITHM);
	log("init: digest algorithm is " + algorithm);
	try {
	    authorization = new Authorization(algorithm);
	}
	catch (java.security.NoSuchAlgorithmException nsae) {
	    log("init: digest algorithm " + algorithm + " missing");
	    return;
	}

	try {			// if this fails, DNS is probably broken
	    me = new IMA(Protocol.SERVER_USER);
	}
	catch (UnknownHostException ue) {
	    log("init: unable to get the local host name");
	    System.exit(1);
	    return;
	}

	key = SimpProperties.NOTIFIER_POOL_LIMIT_KEY;
	int limit = props.getPropertyAsInt(key, 0);
	log("init: notifier pool limit is " + limit);
	notifierPool = new Pool(limit, logger);

	key = SimpProperties.NOTIFIER_PORT_KEY;
	port = props.getPropertyAsInt(key, 0);
	try {
	    notifierSocket = new ServerSocket(port);
	    port = notifierSocket.getLocalPort();
	    log("init: created socket " + me.getHost() + ":" + port);
	    startNotifierThread(notifierSocket);
	}
	catch (IOException ex) {
	    log ("init: failed to find a server socket port", ex);
	    System.exit(1);
	}
    }

    private void startNotifierThread(final ServerSocket ss) {
	Thread t = new Thread(new Runnable() {
		public void run() {
		    try {
			for (;;) {
			    acceptSocket(ss.accept());
			}
		    }
		    catch (IOException ioe) {
			log("notifier thread: I/O exception", ioe);
		    }	  
		    try {
			ss.close();
		    }
		    catch (IOException iox) {
		    }	  
		}
	    });
	t.start();
    }

    /**
     * This routine handles a connection to a client.  The first step
     * is to perform the second half of the login protocol.  If this
     * succeeds, an Endpoint is created and connected to the user's
     * contact place.
     */
    private void acceptSocket(final Socket s) 
	throws IOException
    {
	s.setSoTimeout(SO_TIMEOUT);
	final XpSocket xps = new XpSocket(s);
	xps.addLogListener(logger);
	xps.addXpListener(new XpListener() {
		/**
		 * Handle first XmlProperties object on stream.
		 */
		public void xpReceived(XpEvent e) {
		    try {	
			XmlProperties xp = e.getXp();
			String opaque = xp.getProperty(Protocol.OPAQUE);
			if (opaque == null) {
			    xps.close();
			    return;
			}
			Cookie cookie = cookieTable.remove(opaque);
			if (cookie == null) {
			    String msg = "Authorization failure";
			    xps.write(Protocol.error(msg));
			    log("notifier open: login attempt to " +
				"invalid account");
			    xps.close();
			    return;
			}
			String user = cookie.getUser();
			long now = System.currentTimeMillis();
			if (now > cookie.getDate() + cookie_time_out) {
			    log("notifier open: stale cookie for " + user);
			    xps.close();
			    return;
			}
			String action = xp.getProperty(Protocol.ACTION);
			String auth = xp.getProperty(Protocol.AUTHORIZATION);
			if (auth == null || !Protocol.CONNECT.equals(action)) {
			    log("notifier open: bad connect request by " + 
				user);
			    xps.close();
			    return;
			}
			if (!auth.equals(cookie.getAuth())) {
			    String msg = "Authorization failure";
			    log("notifier open: " + msg + " for " + user);
			    xps.write(Protocol.error(msg));
			    xps.close();
			    return;
			}	
			// user is authenticated
			xps.removeXpListener(this);
			s.setSoTimeout(0);
			ContactPlace place = cpt.get(user);
			place.login(new Endpoint(xps, place));
			XmlProperties prof = place.getProfile();
			if (prof == null) 
			    prof = new XmlProperties();
			prof.setProperty(Protocol.ACTION, Protocol.SELF);
			xps.write(prof);
			place.noteSubscribers();
			log("notifier open: " + user + " connected");
		    }
		    catch (IOException ioe) {
			try {
			    s.close();
			}
			catch (IOException iox) {
			}
			log("notifier open: I/O exception", ioe);
		    }	  
		}

		public void closed() {
		    xps.removeXpListener(this);
		}
	    });
	if (!notifierPool.fork(xps)) {
	    log("notifier open: stream failed due to lack of threads.");
	    try {	// Close socket when server is too busy.
		s.close();
	    }
	    catch (IOException ioe) {
	    }
	}
    }

    /**
     * Routing service.
     */
    XmlProperties service(XmlProperties req) {
	String action = req.getProperty(Protocol.ACTION);
	if (action == null) 
	    return missingAction(req);
	else if (Protocol.LOGIN.equals(action)) 
	    return loginAction(req);
	else {
	    IMA to;
	    try {
		to = new IMA(req.getProperty(Protocol.TO));
	    }
	    catch (UnknownHostException ue) {
		return Protocol.error("To field not an address");
	    }
	    IMA from;
	    try {
		from = new IMA(req.getProperty(Protocol.FROM));
	    }
	    catch (UnknownHostException ue) {
		return Protocol.error("From field not an address");
	    }
	    return localService(to, from, action, req);
	}
    }

    /**
     * Dispatch message to a local contact place.
     */
    private XmlProperties localService(IMA to, IMA from, String action, 
				       XmlProperties req) {
	log("local service: " + action + 
	    " to " + to.toString() +
	    " from " + from.toString());
  	if (!to.isLocal()) {
	    return Protocol.error("Message sent to wrong server");
	}
	if (Protocol.INQUIRE.equals(action)) 
	    return inquire();
	String user = to.getUser();
	ContactPlace cp = cpt.get(user);
	if (cp == null) {
	    log("local service: request of nonexistent user " + user);
	    return Protocol.error("User unknown");
	}
	else if (Protocol.SEND.equals(action)) 
	    return cp.send(to, from, req);
	else if (Protocol.NOTE_CHANGE.equals(action)) 
	    return cp.noteChange(to, from, req);
	else if (Protocol.NOTE_SUBSCRIPTION_END.equals(action)) 
	    return cp.noteSubscriptionEnd(to, from, req);
	else if (Protocol.FETCH.equals(action)) 
	    return cp.fetch(to, from, req);
	else if (Protocol.SUBSCRIBE.equals(action)) 
	    return cp.subscribe(to, from, req);
	else if (Protocol.UNSUBSCRIBE.equals(action)) 
	    return cp.unsubscribe(to, from, req);
	else 
	    return unsupportedAction(action, req);
    }

    XmlProperties missingAction(XmlProperties req) {
	log("request: missing action");
	return Protocol.error("Missing action");
    }

    /**
     * First half of the login protocol.
     */
    private XmlProperties loginAction(XmlProperties req) {
	String user = req.getProperty(Protocol.USER);
	if (user == null) 
	    return Protocol.error("User missing");
	else {
	    String nonce = nonceGen.nextNonce();
	    String opaque = nonceGen.nextNonce();
	    String password = null;
	    try {
		password = pw.getPassword(user);
	    }
	    catch (FileNotFoundException fnfx) {
		log("login service: cannot find password file");
	    }
	    catch (IOException iox) {
		log("login service: error reading password file", iox);
	    }
	    if (password != null) {
		String auth = null;
		try {
		    auth = 
			authorization.getAuthorization(user, 
						       password.toCharArray(), 
						       nonce);
		    // cookie stored for second phase of login protocol
		    cookieTable.put(opaque, new Cookie(user, auth));
		}
		catch (UnsupportedEncodingException uee) {
		    log("login service: unsupported encoding", uee);
		}
		catch (IOException iox) {
		    log("login service: I/O exception", iox);
		}
	    }
	    else
		log("login service: " + user + " has no account");
	    // respond with challenge even when account is bogus
	    return Protocol.challenge(nonce, opaque, 
				      authorization.getAlgorithm(),
				      1, 0, 1, 0, // version range
				      me.getHost(), port);
	}
    }

    private XmlProperties inquire() {
	return Protocol.okay("SIMP Service version " +
			     SimpServer.getMajorVersion() + 
			     "." +
			     SimpServer.getMinorVersion() + 
			     "\nSIMP Protocol version 1.0");
    }

    private XmlProperties unsupportedAction(String action, XmlProperties req) {
	log("request: action " + action + " unsupported");
	return Protocol.error("Action " + action + " unsupported");
    }

    /**
     * Deliver locally generated requests to some routing service.
     */
    XmlProperties deliver(IMA to, IMA from, XmlProperties req) {
	String action = req.getProperty(Protocol.ACTION);
	if (action == null) 
	    return missingAction(req);
	else if (to.isLocal()) 
	    return localService(to, from, action, req);
	try {
	    return HostPost.post(to.getHost(), req);
	}
	catch (IOException ex) {
	    log("deliver: post to " + to + " failed -- " + ex);
	    return Protocol.error("Message could not be delivered");
	}
    }	

    void log(String msg) {
	logger.log(msg);
    }

    void log(String msg, Throwable t) {
	logger.log(msg, t);
    }

    /**
     * This method initiates a shutdown of the server.  Currently,
     * there is no code that invokes this method. 
     */
    void shutdown() {
	// close router server
	try {
	    routerSocket.close();
	}
	catch (IOException iox) {
	}
    }

    /**
     * Reclaim resources on shutdown.
     */
    void destroy() {
	// close notification server
	try {
	    notifierSocket.close();
	}
	catch (IOException iox) {
	}
	// close each endpoint
	cpt.destroy();
	// destroy notification pool
	notifierPool.destroy();
	super.destroy();
    }
}
