// $Id: ACL.java,v 1.10 2000/03/20 11:41:27 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.util;

import java.util.*;
import java.io.*;
import java.net.*;

/**
 * Access control lists for the SIMP system.
 * Permission to perform an operation is decided using a two step
 * process.  The first step maps an IMA to an integer.  The integer
 * encodes the set of allowed operations.  A shift left of one bit
 * position is used when a signed object is involved.
 * @author John D. Ramsdell
 * @version February 2000 
 */
public final class ACL
    extends HashMap
{  
    public final static int OP_SEND = 1<<0; // Operation masks
    public final static int OP_FETCH = 1<<2;
    public final static int OP_SUBSCRIBE = 1<<4;
    public final static int OP_UNSUBSCRIBE = 1<<6;
    public final static int OP_CHANGE = 1<<8;
    public final static int OP_END = 1<<10;

    public final static int OP_NONE = 0;
    public final static int OP_ALL 
	= OP_SEND | OP_SEND << 1
	| OP_FETCH | OP_FETCH << 1
	| OP_SUBSCRIBE | OP_SUBSCRIBE << 1
	| OP_UNSUBSCRIBE | OP_UNSUBSCRIBE << 1
	| OP_CHANGE | OP_CHANGE << 1
	| OP_END | OP_END << 1;

    private int everybody;

    private final XmlProperties xp;

    private ACL(XmlProperties xp) { // object created by encode
	this.xp = xp;
    }

    /** 
     * Decides if an operation is allowed.
     */
    public boolean permit(IMA subject, int operation, boolean validSig) {
	if (validSig)
	    operation <<= 1;	// shift left for signed objects
	Integer val = (Integer)get(subject);
	if (val != null)	// try IMA as key
	    return match(val.intValue(), operation);
	val = (Integer)get(subject.getInetAddress());
	if (val != null)	// try Host as key
	    return match(val.intValue(), operation);
	else			// else use default rule
	    return match(everybody, operation);
    }

    // is bit set?
    private boolean match(int pattern, int operation) {
	return (pattern & operation) == operation;
    }

    /**
     * Returns the allowed set of operations associated with a subject.
     */
    public int getOperations(IMA subject) {
	if (subject == null)
	    return everybody;
	Integer val = (Integer)get(subject);
	if (val != null)	// try subject as key
	    return val.intValue();
	val = (Integer)get(subject.getInetAddress());
	if (val != null)	// try Host as key
	    return val.intValue();
	else			// else use default rule
	    return everybody;
    }

    /**
     * Get the XML Properties object used to create this ACL.
     */
    public XmlProperties getXp() {
	return xp;
    }

    /**
     * Creates an ACL from an XML Properties object. 
     * Returns null on failure.
     */
    public static ACL encode(XmlProperties xp) {
	if (xp == null)
	    return null;
	ACL acl = new ACL(xp);
	String all = xp.getProperty(Protocol.EVERYBODY);
	acl.everybody = all != null ? encodeOps(all) : OP_ALL;
	if (acl.everybody < 0)
	    return null;
	for (Enumeration e = xp.keys(); e.hasMoreElements();) {
	    String key = (String)e.nextElement();
	    if (Protocol.EVERYBODY.equals(key))
		continue;
	    if (!IMA.isIMA(key))
		return null;
	    try {
		IMA subject = new IMA(key);
		int ops = encodeOps(xp.getProperty(key));
		if (ops < 0)
		    return null;
		if ("".equals(subject.getUser())) {
		    if (null != acl.put(subject.getInetAddress(),
					new Integer(ops)))
			return null;
		}
		else if (null != acl.put(subject, new Integer(ops)))
		    return null;
	    }
	    catch (UnknownHostException ue) {
		return null;
	    }
	}
	return acl;
    }

    // maps a list of operations to an integer encoding.
    private static int encodeOps(String s) {
	if (s == null)
	    return -1;
	int mask = 0;
	StringTokenizer st = new StringTokenizer(s);
	while (st.hasMoreTokens()) {
	    int op = 0;
	    String t = st.nextToken();
	    boolean mustSign = t.charAt(0) == '+';
	    if (mustSign)
		t = t.substring(1);
	    if (Protocol.SEND.equals(t))
		op = OP_SEND;
	    else if (Protocol.FETCH.equals(t))
		op = OP_FETCH;
	    else if (Protocol.SUBSCRIBE.equals(t))
		op = OP_SUBSCRIBE;
	    else if (Protocol.UNSUBSCRIBE.equals(t))
		op = OP_UNSUBSCRIBE;
	    else if (Protocol.CHANGE.equals(t))
		op = OP_CHANGE;
	    else if (Protocol.END.equals(t))
		op = OP_END;
	    else {
		return -1;
	    }
	    if (!mustSign)
		mask |= op;
	    op <<= 1;
	    mask |= op;
	}
	return mask;
    }
}
