// $Id: Base64.java,v 1.3 2000/01/26 11:36:44 ramsdell Exp $

/* Base64 encoding and decoding functions written in Java
Copyright (C) 1998 Nathan C. Summers

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

As a special exception to the LGPL, you may change the package in 
which this class resides without being required to document and 
distribute the modifications as long as the package location is the
only change made.
*/

// package loki.util;
package org.mitre.simp.util;

/**
 * A class that implements encoding and decoding Base64.
 *
 * I should get around to making these filterstreams.
 *
 * License: LGPL
 *
 * @author Rockwalrus
 */

public final class Base64 {
	private final static char[] alphabet = 
		{'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 
		 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
		 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
		 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
		 '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/'};

	/**
	 * Base64 encodes the input.
	 * 
	 * @param msg the message to be encoded
	 * @return the base64 of msg.  Note that mail applications may need to inset a CRLF every 76 chars to be smtp compliant.
	 */
	public static char[] encode (byte[] msg) {
		char[] ret = new char[4 * (msg.length + ((3 - (msg.length % 3)) % 3)) / 3];
		int r=0, i=0, left, bitqueue;

		while (i < msg.length) {
			left = msg.length - i;

			if (left > 2) {
				bitqueue = msg[i++];
				bitqueue = (bitqueue << 8) + msg[i++];
				bitqueue = (bitqueue << 8) + msg[i++];

				ret[r++] = alphabet[(bitqueue & 0xFC0000) >> 18];
				ret[r++] = alphabet[(bitqueue &  0x3F000) >> 12];
				ret[r++] = alphabet[(bitqueue &    0xFC0) >>  6];
				ret[r++] = alphabet[(bitqueue &     0x3F)      ];
			} else if (left == 2) {
				bitqueue = msg[i++];
				bitqueue = (bitqueue << 8) + msg[i++];
				bitqueue <<= 8;

				ret[r++] = alphabet[(bitqueue & 0xFC0000) >> 18];
				ret[r++] = alphabet[(bitqueue &  0x3F000) >> 12];
				ret[r++] = alphabet[(bitqueue &    0xFC0) >>  6];
				ret[r++] = '=';
			} else {
				bitqueue = msg[i++];
				bitqueue <<= 16;

				ret[r++] = alphabet[(bitqueue & 0xFC0000) >> 18];
				ret[r++] = alphabet[(bitqueue &  0x3F000) >> 12];
				ret[r++] = '=';
				ret[r++] = '=';
			}
		}
		
		return ret;
	}

	/**
	 * Base64 decodes the input.
	 * 
	 * @param msg the message to be decoded.  Note that all whitespace and line breaks should be removed before calling.
	 * @return the original message.
	 * @exception loki.util.Base64Exception if msg is not a legal base64 message.
	 */
	public static byte[] decode (char[] msg) throws Base64Exception {
		if ((msg.length % 4) != 0) 
			throw (new Base64Exception("Message length not multiple of 4."));

		int retlength = msg.length / 4 * 3;

		if (msg[msg.length - 1] == '=') {
			retlength--;
			if (msg[msg.length - 2] == '=')
				retlength--;
		}
			
		byte[] ret = new byte[retlength];

		int i=0, r=0, bitqueue;
		while (i < msg.length) {
			int pad = 0;
			byte tmp = unalphabet(msg[i++]);

			if (tmp == 64)
				throw (new Base64Exception("Illegal padding at index " + --i));
			bitqueue = tmp;
	
			tmp = unalphabet(msg[i++]);
			if (tmp == 64)
				throw (new Base64Exception("Illegal padding at index " + --i));
			bitqueue <<= 6;
			bitqueue += tmp;

			tmp = unalphabet(msg[i++]);
			if (tmp == 64) {
				if (i != msg.length - 1)
					throw (new Base64Exception("Illegal padding at index " + --i));
				tmp = 0; pad++;
			}

			bitqueue <<= 6;
			bitqueue += tmp;
			tmp = unalphabet(msg[i++]);
			if (pad == 1 && tmp != 64)
				throw (new Base64Exception("Charactor after padding at index " + --i));

			if (tmp == 64) {
				if (i != msg.length)
					throw (new Base64Exception("Illegal padding at index " + --i));
				tmp = 0; pad++;
			}
			
			bitqueue <<= 6;
			bitqueue += tmp;
			
			ret[r++] = (byte) ((bitqueue & 0xFF0000) >> 16);
			if (pad < 2) {
				ret[r++] = (byte) ((bitqueue & 0xFF00) >> 8);
				if (pad < 1)
					ret[r++] = (byte) (bitqueue & 0xFF);
			}
		}

		return ret;
	}

	private static byte unalphabet(char msg) throws Base64Exception {
		if (msg >= 'A' && msg <= 'Z')
			return (byte) (msg - 'A');
		else if (msg >= 'a' && msg <= 'z')
			return (byte) (msg - 'a' + 26);
		else if (msg >= '0' && msg <= '9')
			return (byte) (msg - '0' + 52);
		else if (msg == '+')
			return 62;
		else if (msg == '/')
			return 63;
		else if (msg == '=')
			return 64;
		else 
			throw (new Base64Exception("Invalid charactor: " + msg));
	}
}
