// $Id: RunQueue.java,v 1.4 2000/03/24 14:14:29 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

// A queue for runnable tasks

package org.mitre.simp.util;

import java.util.Vector;

/**
 * A queue for runnable tasks.
 * The run method sequentially invokes tasks in the queue.
 * The run queue is meant to be run in its own thread.
 * @version January 1998
 * @author John D. Ramsdell
 */
public class RunQueue
    extends LogGenerator
    implements Runnable
{
    // A queue of runnables
    private final Vector queue = new Vector();

    // Indicates that tasks extracted from the queue should not be started
    private volatile boolean flushQueue = false;

    /**
     * Add a task to the run queue.
     * @param task a runnable task
     */
    public synchronized void add(Runnable task) {
	queue.addElement(task);
	notify();
    }

    /**
     * Remove all tasks from the run queue.
     */
    public synchronized void removeAll() {
	flushQueue = true;		// Stop all tasks that 
	queue.removeAllElements();	// have not been started
    }

    /**
     * Runs tasks in a task queue.
     */
    public void run() {
	Vector q;
	while (!Thread.interrupted()) {
	    try {
		synchronized (this) {
		    while (queue.isEmpty())
			wait();
		    q = (Vector)queue.clone(); // Extract tasks from the queue
		    queue.removeAllElements();
		    flushQueue = false;
		}
	    }
	    catch (InterruptedException ie) {
		break;			// Thread was interrupted
	    }
	    // Abort loop when flushQueue is enabled
	    for (int i = 0; i < q.size() && !flushQueue; i++) {
		Runnable task = (Runnable)q.elementAt(i);
		try {
		    task.run();
		}
		catch (Throwable t) {
		    log("Internal error: " + t.getMessage(), t);
		}
	    }
	    q = null;			// Recycle queue.
	}
    }

    /**
     * Generate a LogEvent for any registered LogListeners.
     */
    public void log(String msg, Throwable t) {
	super.log("run queue: " + msg, t);
    }
}
