// $Id: SocketPost.java,v 1.8 2000/02/29 20:29:54 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.util;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.util.*;

/**
 * Post via an XML Properties socket.  A socket post can be used to
 * issue a request and wait for a reply.  More than one thread can
 * make use of a socket post at the same time.  Replies need not
 * arrive in the order in which the corresponding requests were
 * issued.
 * @author John D. Ramsdell
 * @version February 2000
 */
public class SocketPost
    implements XpListener
{
    private final static long TIME_OUT = 30 * 1000;

    public final static String REQUEST_KEY = "request";
    public final static String RESPONSE_KEY = "response";

    private final XpSocket socket;

    /** 
     * Create a socket post on top of an XML Properties socket
     */
    public SocketPost(XpSocket socket)
    {
	this.socket = socket;
	socket.addXpListener(this);
    }

    public XpSocket getXpSocket() {
	return socket;
    }

    private String getId() {
	return Long.toString(socket.nextSeqNo());
    }

    private final Hashtable ht = new Hashtable();

    /**
     * Post a request via the underlying XML Properties socket and
     * return the reply.  
     */
    public XmlProperties post(XmlProperties req)
	throws IOException
    {
	long done = TIME_OUT + System.currentTimeMillis();
	String id = getId();
	req.setProperty(REQUEST_KEY, id);
	socket.write(req);
	try {
	    synchronized (ht) {
		for (;;) {
		    long now = System.currentTimeMillis();
		    XmlProperties xp = (XmlProperties)ht.remove(id);
		    if (xp != null) 
			return xp;
		    if (now < done)
			ht.wait(done - now);
		    else
			throw new IOException("Time out in socket post read");
		}
	    }
	}
	catch (InterruptedException ie) {
	    throw new InterruptedIOException("Socket post read interrupted");
	}
    }

    public void closed() {
    }

    public void xpReceived(XpEvent e) {
	XmlProperties xp = e.getXp();
	String id = xp.remove(RESPONSE_KEY);
	if (id != null) {
	    synchronized (ht) {
		ht.put(id, xp);
		ht.notifyAll();
	    }
	}
    }

    /**
     * Write to the underlying XML Properties socket.
     */
    public void write(XmlProperties xp) 
	throws IOException
    {
	socket.write(xp);
    }

    /**
     * Close the underlying XML Properties socket.
     */
     public void close() 
	throws IOException
    {
	socket.close();
    }
}
