// $Id: XmlProperties.java,v 1.12 2000/03/01 15:43:16 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

// Properties using XML as the external format.

package org.mitre.simp.util;

import java.util.Hashtable;
import java.util.Enumeration;
import java.io.*;
import org.xml.sax.helpers.ParserFactory;
import org.xml.sax.*;

/** 
 * This class implements a persistent hashtable that maps strings to
 * strings.  An XML Properties object can be saved to a stream or
 * loaded from a stream.  XML is used as the external format.
 * @see java.util.Properties
 * @author John D. Ramsdell
 * @version February 2000
 */
public class XmlProperties
{
    private final Hashtable ht;

    public XmlProperties(int initialCapacity, float loadFactor) {
	ht = new Hashtable(initialCapacity, loadFactor);
    }

    public XmlProperties(int initialCapacity) {
	ht = new Hashtable(initialCapacity);
    }

    /**
     * Create an XML Properties object.
     */
    public XmlProperties() {
	ht = new Hashtable();
    }

    private XmlProperties(Hashtable ht) { // For clone
	this.ht = ht;
    }

    public Object clone() {
	return new XmlProperties((Hashtable)ht.clone());
    }

    /**
     * Returns the string associated with the key, otherwise null.
     */
    public String getProperty(String key) {
	return (String)ht.get(key);
    }

    /**
     * Associates the value with the key.
     */
    public String setProperty(String key, String value) {
	return (String)ht.put(key, value);
    }

    /**
     * Removes the association of a value with the key.
     * The previously associated value is returned, 
     * or null if there is none.
     */
    public String remove(String key) {
	return (String)ht.remove(key);
    }

    /**
     * Returns the string associated with the key, otherwise it
     * returns the default value.
     */
    public String getProperty(String key, String defaultValue) {
	Object value = ht.get(key);
	if (value == null)
	    return defaultValue;
	else 
	    return (String)value;
    }

    /**
     * Get an integer.
     * @param key name of key
     * @param defaultValue value on error
     * @return property as an integer or default value on error
     */
    public int getPropertyAsInt(String key, int defaultValue) {
	String str = getProperty(key);
	if (str == null)
	    return defaultValue;
	try {
	    return Integer.parseInt(str);
	}
	catch (NumberFormatException ex) {
	    return defaultValue;
	}
    }

    /**
     * Get a long integer.
     * @param key name of key
     * @param defaultValue value on error
     * @return property as a long integer or default value on error
     */
    public long getPropertyAsLong(String key, long defaultValue) {
	String str = getProperty(key);
	if (str == null)
	    return defaultValue;
	try {
	    return Long.parseLong(str);
	}
	catch (NumberFormatException ex) {
	    return defaultValue;
	}
    }

    /**
     * Get a boolean.
     * @param key name of key
     * @return property as a boolean or the false value on error
     */
    public boolean getPropertyAsBoolean(String key) {
	String str = getProperty(key);
	return Boolean.valueOf(str).booleanValue();
    }

    /**
     * Returns the number of entries in this properties object.
     */
    public int size() {
	return ht.size();
    }

    /**
     * Returns an enumeration of the keys in this object.
     */
    public Enumeration keys() {
	return ht.keys();
    }

    // Loader

    private final static String SAX_PARSER_CLASS = "com.jclark.xml.sax.Driver";
    private final static String PROPERTIES_TAG = "properties";
    private final static String HEADER_TAG = "header";
    private final static String ENTRY_TAG = "entry";
    private final static String KEY_TAG = "key";

    /**
     * Returns the SAX parser used when reading the object's 
     * external format.
     */
    public static String getSAXParserClass() {
	return SAX_PARSER_CLASS;
    }

    private DocumentHandler handler;

    /**
     * Reads an XML properties object from an input stream. 
     * @param      in   the input stream.
     * @exception  IOException  if an error occurred when reading from the
     *               input stream.
     */
    public void load(InputStream in)
	throws IOException
    {
	try {
	    Parser parser = ParserFactory.makeParser(SAX_PARSER_CLASS);
	    if (handler == null)
		handler = new LoaderHandler();
	    parser.setDocumentHandler(handler);
	    parser.parse(new InputSource(in));
	}
	catch (ClassNotFoundException cnfe) {
	    throw new IOException("SAX Parser class not found: " + 
				  SAX_PARSER_CLASS);
	}
	catch (IllegalAccessException iae) {
	    throw new IOException("Permission to load class denied: " +
				  SAX_PARSER_CLASS);
	}
	catch (InstantiationException ie) {
	    throw new IOException("Cannot instantiate: " +
				  SAX_PARSER_CLASS);
	}
	catch (ClassCastException cce) {
	    String msg = "Class does not implement org.xml.sax.Parser: " +
		SAX_PARSER_CLASS;
	    throw new IOException(msg);
	}
	catch (SAXException se) {
	    throw new IOException("SAX Exception: " + se);
	}
    }

    /**
     * Reads an XML properties object from a string.
     * @param      s  the string.
     * @exception  IOException  if an error occurred when reading from the
     *               string.
     */
    public void load(String s)
	throws IOException
    {
	load(new ByteArrayInputStream(s.getBytes("UTF8")));
    }

    private class LoaderHandler
	implements DocumentHandler
    {
	// Loader state variables
	private boolean properties;	// properties element seen
	private String key;		// Non-null in an entry element
	private StringBuffer sb;	// Non-null in an entry element
	private Locator locator;

	// Application methods

	public void setDocumentLocator(Locator locator) {
	    this.locator = locator;
	}

	private void reportParseError(String message) 
	    throws SAXParseException
	{
	    throw new SAXParseException(message, locator);
	}

	public void startDocument() {
	    properties = false;
	    key = null;
	}

	public void endDocument() { }

	public void startElement(String name, AttributeList attrs) 
	    throws SAXParseException
	{
	    if (key != null)
		reportParseError("element " + name + 
				 " not valid within an entry");
	    if (ENTRY_TAG.equals(name)) {
		if (!properties) 
		    reportParseError(ENTRY_TAG + " element misplaced");
		key = attrs.getValue(KEY_TAG);
		if (key == null)
		    reportParseError(ENTRY_TAG + " " + KEY_TAG +
				     " attribute missing");
		sb = new StringBuffer();
	    
	    }
	    else if (PROPERTIES_TAG.equals(name)) {
		if (properties) 
		    reportParseError("Input not valid");
		properties = true;
	    }
	    else
		reportParseError("element " + name + " not valid");
	}

	public void characters(char ch[], int start, int length) {
	    if (sb != null) {
		sb.append(ch, start, length);
	    }
	}

	public void endElement(String name) {
	    if (PROPERTIES_TAG.equals(name))
		properties = false;
	    else if (ENTRY_TAG.equals(name) && key != null) 
		setProperty(key, sb.toString());
	    key = null;
	    sb = null;
	}

	public void ignorableWhitespace(char[] ch, int start, int length) {}
	public void processingInstruction(String target, String data) {}
    }

    // Store

    /**
     * Stores this properties object to the specified output stream.
     * If non-null, the string header is printed as a header attribute
     * of the properties.
     * @param   out      an output stream.
     * @param   header   a description of the property list.
     * @exception  IOException  if an error occurred while writing to the
     *               output stream.  
     */
    public void store(OutputStream out, String header) 
	throws IOException
    {
	UTF8XMLWriter w = new UTF8XMLWriter(out);
	w.markup(prolog);
	w.markup("<" + PROPERTIES_TAG);
	if (header != null) {
	    w.markup(" " + HEADER_TAG + "='");
	    w.write(header);
	    w.markup("'");
	}
	w.markup(">\n");

	for (Enumeration e = keys(); e.hasMoreElements();) {
	    String key = (String)e.nextElement();
	    String value = getProperty(key);
	    w.markup("<" + ENTRY_TAG + " " + KEY_TAG + "='");
	    w.write(key);
	    w.markup("'>");
	    w.write(value);
	    w.markup("</" + ENTRY_TAG + ">\n");
	}

	w.markup("</" + PROPERTIES_TAG + ">\n");
	w.flush();
    }

    /**
     * Produces a String containing the external
     * representation of the properties object as an XML document.
     * @return the external representation as a String.  
     * @exception  IOException  if an error occurred while creating
     * 			the string.
     */
    public String asString() 
	throws IOException
    {
	ByteArrayOutputStream uout = new ByteArrayOutputStream();
	store(uout, null);
	return uout.toString("UTF8");
    }

    private final static String prolog
	= "<?xml version=\"1.0\"?>\n"
	+ "<!DOCTYPE properties [\n"
	+ "  <!ELEMENT properties (entry)*>\n"
	+ "  <!ATTLIST properties header CDATA #IMPLIED>\n"
	+ "  <!ELEMENT entry (#PCDATA)>\n"
	+ "  <!ATTLIST entry key CDATA #REQUIRED>\n"
	+ "]>\n";

    /**
     * A routine for testing.
     * It loads some key value pairs and then loads properties 
     * in the files given in args.
     */
    public static void main(String[] args) 
    {
	XmlProperties xp = new XmlProperties();

	xp.setProperty("key", "value");
	xp.setProperty("lt gt amp quot apos", "<>&\"'");
	xp.setProperty("<<angle>>", "\"quote&quote\"");

	try {
	    for (int i = 0; i < args.length; i++) {
		xp.load(new FileInputStream(args[i]));
	    }
	    xp.store(System.out, "a `little' header & more");
	}
	catch (Throwable t) {
	    System.err.println(xp.getClass().getName()
			       + " caught an exception");
	    t.printStackTrace();
	}
    }
}
