// $Id: XpSocket.java,v 1.17 2000/03/24 14:14:29 ramsdell Exp $

// Copyright (c) 2000. The MITRE Corporation (http://www.mitre.org/).
// All rights reserved.
// The SIMP Service comes with ABSOLUTELY NO WARRANTY.
// See the SIMP Service License Agreement for details.

package org.mitre.simp.util;

import java.io.*;
import java.net.*;
import java.util.*;
import org.xml.sax.helpers.ParserFactory;
import org.xml.sax.*;

/**
 * Sockets that pass a stream of XML Properties objects.  
 * A write of an XML Properties object causes output to the stream.
 * Input is event based.  Objects interested in input must register an
 * XpListener.  The thread running the XML Properties socket's run
 * method generates the events.
 * @author John D. Ramsdell
 * @version February 2000 
 */
public class XpSocket
    extends LogGenerator
    implements Runnable
{
    // Turn this on to see writes to this object
    private final static boolean DEBUG = false;

    // Use this unless you are using a version of Java that supports
    // half closed sockets (You have to edit other parts to turn this off).
    private final static boolean USE_UGLY_SOCKET_HACK = true;

    // An XmlProperties document is dropped when the following
    // limits are exceeded.
    private final static int KEY_COUNT_LIMIT = 50;
    private final static int VALUE_SIZE_LIMIT = 1 << 13;

    // Tags used in documents
    private final static String STREAM_TAG = "stream";
    private final static String PROPERTIES_TAG = "properties";
    private final static String ENTRY_TAG = "entry";
    private final static String KEY_TAG = "key";

    private Socket socket;
    private BufferedInputStream in;
    private UTF8XMLWriter out;

    /**
     * Create an XML Properties socket stream that passes objects
     * over the given socket.
     */
    public XpSocket(Socket socket) 
	throws IOException
    {
	this.socket = socket;
	in = new BufferedInputStream(socket.getInputStream());
	OutputStream os = socket.getOutputStream();
	out = new UTF8XMLWriter(new BufferedOutputStream(os));
    }

    // Output

    private boolean opened = false;

    private final static String prolog
	= "<?xml version=\"1.0\"?>\n"
	+ "<" + STREAM_TAG + ">\n";

    /** 
     * Write an XML Properties object to the stream.
     */
    public synchronized void write(XmlProperties xp)  
	throws IOException
    {
	if (!opened) {
	    out.markup(prolog);
	    opened = true;
	}

	out.markup("<" + PROPERTIES_TAG + ">\n");

	for (Enumeration e = xp.keys(); e.hasMoreElements();) {
	    String key = (String)e.nextElement();
	    String value = xp.getProperty(key);
	    out.markup("<" + ENTRY_TAG + " " + KEY_TAG + "='");
	    out.write(key);
	    out.markup("'>");
	    out.write(value);
	    out.markup("</" + ENTRY_TAG + ">\n");
	}

	out.markup("</" + PROPERTIES_TAG + ">\n");
	out.flush();
	if (DEBUG)
	    xp.store(System.err, XpSocket.class.getName());
    }

    private boolean closed = false;

    /**
     * Close a stream for output.
     */
    public synchronized void close()
	throws IOException
    {
	if (closed) 
	    return;
	else
	    closed = true;
	try {
	    if (!opened) {
		out.markup(prolog);
		opened = true;
	    }
	    out.markup("</" + STREAM_TAG + ">\n");
	    out.flush();
	}
	finally {
	    try {
		out.close();
	    }
	    catch (IOException iox) { }
	    /* What I really want is:
	       if (!USE_UGLY_SOCKET_HACK) {
	       try {
	       socket.shutdownOutput();
	       }
	       catch (IOException iox) { }
	       }
	    */
	}
    }

    protected void finalize() 
	throws IOException
    {
	close();
    }

    /**
     * Close without notification sent to any LogListeners.
     */
    public void kill()
	throws IOException
    {
	listeners.clear();
	close();
    }

    /** 
     * Input is read from the socket by this method and given to
     * registered XpListeners.  
     */
    public void run() {
	String parserClass = XmlProperties.getSAXParserClass();
	try {
	    Parser parser = ParserFactory.makeParser(parserClass);
	    parser.setDocumentHandler(new XpHandler());
	    parser.parse(new InputSource(in));
	}
	catch (ClassNotFoundException cnfe) {
	    log("SAX parser " + parserClass + " not found", cnfe);
	}
	catch (IllegalAccessException iae) {
	    log("permission to load " + parserClass + " denied", iae);
	}
	catch (InstantiationException ie) {
	    log("cannot instantiate " + parserClass, ie);
	}
	catch (ClassCastException cce) {
	    log(parserClass + " does not implement org.xml.sax.Parser", cce);
	}
	catch (SAXException se) {
	    Exception ex = se.getException();
	    if (!(ex instanceof EndOfInputException))
		log(se.getMessage(), ex);
	}
	catch (InterruptedIOException ioe) {
	    log("time out during read");
	}
	catch (SocketException ioe) {
	    log(ioe.toString(), null);
	}
	catch (IOException ioe) {
	    log("I/O exception during read: " + ioe, ioe);
	}
	catch (Throwable t) {
	    log("internal error during read: " + t, t);
	}
	finally {
	    try {
		close();
	    }
	    catch (IOException iox) {
		// log("I/O exception during close: " + iox, iox);
	    }
	    notifyOfClose();
	    try {
		socket.close();
	    }
	    catch (IOException iox) {
		// Ignore these errors
	    }
	}
    }

    // The parser handler could be implemented so that it creates less
    // objects, however, this implementation is easier to understand.

    private class XpHandler
	implements DocumentHandler
    {
	private DocumentHandler handler = new Start();
	private Locator locator;

	private void setHandler(DocumentHandler handler) {
	    this.handler = handler;
	}

	public void setDocumentLocator(Locator locator) {
	    this.locator = locator;
	}

	public void startDocument()
	    throws SAXException
	{
	    handler.startDocument();
	}

	public void endDocument()
	    throws SAXException
	{
	    handler.endDocument();
	}

	public void startElement(String name, AttributeList attrs)
	    throws SAXException
	{
	    handler.startElement(name, attrs);
	}

	public void endElement(String name)
	    throws SAXException
	{
	    handler.endElement(name);
	}

	public void characters(char ch[], int start, int length)
	    throws SAXException
	{
	    handler.characters(ch, start, length);
	}

	public void ignorableWhitespace(char ch[], int start, int length)
	    throws SAXException
	{
	    handler.ignorableWhitespace(ch, start, length);
	}

	public void processingInstruction(String target, String data)
	    throws SAXException
	{
	    handler.processingInstruction(target, data);
	}

	private void reportParseError(String message) 
	    throws SAXParseException
	{
	    throw new SAXParseException(message, locator);
	}

	// ******** Parser states:

	private class Base
	    implements DocumentHandler
	{
	    public void setDocumentLocator(Locator locator) {
	    }

	    public void startDocument()
		throws SAXException
	    {
		reportParseError("document start not expected here");
	    }

	    public void endDocument()
		throws SAXException
	    {
		reportParseError("document end not expected here");
	    }

	    public void startElement(String name, AttributeList attrs)
		throws SAXException
	    {
		reportParseError(name + " element start not expected here");
	    }

	    public void endElement(String name)
		throws SAXException
	    {
		reportParseError(name + " element end not expected here");
	    }

	    public void characters(char ch[], int start, int length)
		throws SAXException
	    {
	    }

	    public void ignorableWhitespace(char ch[], int start, int length) {
	    }

	    public void processingInstruction(String target, String data)
		throws SAXException 
	    {
		reportParseError("processing instruction not expected");
	    }
	}
  
	// Start state
	private class Start
	    extends Base
	{
	    public void startDocument() {
		setHandler(new AfterStart());
	    }
	}    

	private class AfterStart
	    extends Base
	{
	    public void startElement(String name, AttributeList attrs)
		throws SAXException
	    {
		if (STREAM_TAG.equals(name)) {
		    setHandler(new GetProperties(this));
		}
		else
		    reportParseError("expecting " + STREAM_TAG +
				     " element start but found " + name);
	    }

	    public void endDocument() { 
		// We are done with no errors
	    }
	}

	private class GetProperties
	    extends Base
	{
	    private final DocumentHandler stk;

	    GetProperties(DocumentHandler stk) {
		this.stk = stk;
	    }

	    public void startElement(String name, AttributeList attrs)
		throws SAXException
	    {
		if (PROPERTIES_TAG.equals(name)) {
		    setHandler(new GetEntries(this));
		}
		else
		    reportParseError("expecting " + PROPERTIES_TAG +
				     " element start but found " + name);
	    }

	    public void endElement(String name)
		throws SAXException
	    {
		if (STREAM_TAG.equals(name)) {
		    setHandler(stk);
		    if (USE_UGLY_SOCKET_HACK) {
			Exception ex = new EndOfInputException();
			throw new SAXException("done", ex);
		    }
		}
		else
		    reportParseError("expecting " + STREAM_TAG +
				     " element end but found " + name);
	    }
	}

	private class GetEntries
	    extends Base
	{
	    private final DocumentHandler stk;
	    private final XmlProperties xp = new XmlProperties();

	    GetEntries(DocumentHandler stk) {
		this.stk = stk;
	    }

	    public void startElement(String name, AttributeList attrs)
		throws SAXException
	    {
		if (ENTRY_TAG.equals(name)) {
		    String key = attrs.getValue(KEY_TAG);
		    if (key == null)
			throw new SAXException("entry key attribute missing");
		    if (xp.size() > KEY_COUNT_LIMIT) {
			setHandler(new BadValue(new BadEntries(stk)));
			log("properties dropped due to too many keys", null);
		    }
		    else
			setHandler(new GetValue(this, xp, key));
		}
		else
		    reportParseError("expecting " + ENTRY_TAG +
				     " element start but found " + name);
	    }

	    public void endElement(String name)
		throws SAXException
	    {
		if (PROPERTIES_TAG.equals(name)) {
		    notifyOfXpReceived(xp);
		    setHandler(stk);
		}
		else
		    reportParseError("expecting " + PROPERTIES_TAG +
				     " element end but found " + name);
	    }
	}
  
	private class GetValue
	    extends Base
	{
	    private final DocumentHandler stk;
	    private final XmlProperties xp;
	    private final String key;
	    private final StringBuffer sb = new StringBuffer();

	    GetValue(DocumentHandler stk, XmlProperties xp, String key) {
		this.stk = stk;
		this.xp = xp;
		this.key = key;
	    }

	    public void characters(char ch[], int start, int length) {
		if (length + sb.length() > VALUE_SIZE_LIMIT) {
		    setHandler(new BadValue(stk));
		    log("property value for " + KEY_TAG + "='" +
			key + "' dropped due to large size", null);
		}
		else
		    sb.append(ch, start, length);
	    }

	    public void endElement(String name)
		throws SAXException
	    {
		if (ENTRY_TAG.equals(name)) {
		    xp.setProperty(key, sb.toString());
		    setHandler(stk);
		}
		else
		    reportParseError("expecting " + ENTRY_TAG +
				     " element end but found " + name);
	    }
	}

	private class BadEntries
	    extends Base
	{
	    private final DocumentHandler stk;

	    BadEntries(DocumentHandler stk) {
		this.stk = stk;
	    }

	    public void startElement(String name, AttributeList attrs)
		throws SAXException
	    {
		if (ENTRY_TAG.equals(name)) {
		    String key = attrs.getValue(KEY_TAG);
		    if (key == null)
			throw new SAXException("entry key attribute missing");
		    setHandler(new BadValue(this));
		}
		else
		    reportParseError("expecting " + ENTRY_TAG +
				     " element start but found " + name);
	    }

	    public void endElement(String name)
		throws SAXException
	    {
		if (PROPERTIES_TAG.equals(name)) {
		    setHandler(stk);
		}
		else
		    reportParseError("expecting " + PROPERTIES_TAG +
				     " element end but found " + name);
	    }
	}
  
	private class BadValue
	    extends Base
	{
	    private final DocumentHandler stk;

	    BadValue(DocumentHandler stk) {
		this.stk = stk;
	    }

	    public void characters(char ch[], int start, int length) {
	    }

	    public void endElement(String name)
		throws SAXException
	    {
		if (ENTRY_TAG.equals(name)) {
		    setHandler(stk);
		}
		else
		    reportParseError("expecting " + ENTRY_TAG +
				     " element end but found " + name);
	    }
	}
    }

    private static class EndOfInputException
	extends Exception
    {
    }

    // Sequence numbers for SocketPost

    // Sequence numbers must be unique within each instance
    // of an XpSocket.
    private static class Counter 
    {
	long seqNo = 0;

	synchronized long nextSeqNo() {
	    return seqNo++;
	}
    }

    private Counter counter = new Counter();

    long nextSeqNo() {
	return counter.nextSeqNo();
    }

    // Xp listener

    private Vector listeners = new Vector();

    public synchronized void addXpListener(XpListener xpl) {
	listeners.add(xpl);
    }

    public synchronized void removeXpListener(XpListener xpl) {
	listeners.remove(xpl);
    }

    private void notifyOfXpReceived(XmlProperties xp) {
	Vector l;
	XpEvent e = new XpEvent(this, xp);
    
	synchronized (this) { l = (Vector)listeners.clone(); }

	for (int i = 0; i < l.size(); i++) {
	    ((XpListener)l.elementAt(i)).xpReceived(e);
	}
    }

    private void notifyOfClose() {
	Vector l;
    
	synchronized (this) { l = (Vector)listeners.clone(); }

	for (int i = 0; i < l.size(); i++) {
	    ((XpListener)l.elementAt(i)).closed();
	}
    }

    /**
     * Generate a LogEvent for any registered LogListeners.
     */
    public void log(String msg, Throwable t) {
	super.log("xpsocket: " + msg, t);
    }
}  
